# from psycopg2.errors import DuplicateDatabase, InvalidCatalogName
import psycopg.errors

from .psyco import quoted_identifier as qi

NO_YES_REALLY = (
    "You tried to drop a database without setting `yes_really_drop` to true."
)


class DatabaseCreateDrop:
    def check_connection(self):
        return self.q("select 1")

    def can_connect(self):
        try:
            self.check_connection()
            return True
        except Exception:
            return False

    def create_db(
        self, name, *, fail_if_exists=False, drop_if_exists=False, yes_really_drop=False
    ):
        _name = qi(name)

        recreate = False

        if drop_if_exists and fail_if_exists:
            raise ValueError("Cannot both drop_if_exists and fail_if_exists")

        try:
            self.autocommit(f"create database {_name};")

        except psycopg.errors.DuplicateDatabase:
            if fail_if_exists:
                raise

            elif drop_if_exists:
                if not yes_really_drop:
                    raise RuntimeError(NO_YES_REALLY)
                self.drop_db(name, yes_really_drop=yes_really_drop)
                recreate = True

        if recreate:
            self.autocommit(f"create database {_name};")

        return self.sibling(name)

    def drop_db(
        self, name, *, fail_if_not_exists=False, yes_really_drop=False, force=False
    ):
        _name = qi(name)

        if not fail_if_not_exists:
            ifexist = "if exists "
        else:
            ifexist = ""

        if force:
            forceclause = " with (force)"
        else:
            forceclause = ""

        try:
            if not yes_really_drop:
                raise RuntimeError(NO_YES_REALLY)
            self.autocommit(f"drop database {ifexist}{_name}{forceclause};")
        except psycopg.errors.InvalidCatalogName:
            if fail_if_not_exists:
                raise



# Copyright results Pty Ltd 2022

import sys

import click

import results


@click.group()
def cli():
    pass


def init_commands(cli):
    @cli.command(help="`diff` two databases, a -> b")
    @click.option("--schema", help="Restrict output to single schema", default=None)
    @click.option(
        "--exclude-schema",
        help="Restrict output to statements for all schemas except the specified schema",
        default=None,
    )
    @click.option(
        "--create-extensions-only",
        is_flag=True,
        help='Only output "create extension..." statements, nothing else',
    )
    @click.option(
        "--ignore-extension-versions",
        is_flag=True,
        help="Ignore the versions when comparing extensions",
    )
    @click.option(
        "--with-privileges",
        is_flag=True,
        default=False,
        help="Also output privilege differences (ie. grant/revoke statements)",
    )
    @click.argument("db_url_a", type=str, nargs=1)
    @click.argument("db_url_b", type=str, nargs=1)
    def dbdiff(db_url_a, db_url_b, **kwargs):
        db_a = results.db(db_url_a)
        db_b = results.db(db_url_b)

        schemadiff_sql = db_a.schemadiff_as_sql(db_b, **kwargs)

        if schemadiff_sql:
            print(schemadiff_sql)
            sys.exit(2)


init_commands(cli)



PROMPT_DEFAULT = (
    "Are you sure you wish to apply the statements above to your database?\n"
    "Type 'yes' to apply them, or any other answer to cancel: "
)
MATCHING_MESSAGE = "Schemas match, no changes required.\n"


class Schemas:
    def schemadiff_as_statements(
        self,
        other,
        schema=None,
        exclude_schema=None,
        create_extensions_only=False,
        ignore_extension_versions=False,
        with_privileges=False,
    ):
        from results.dbdiff import Migration

        m = Migration(
            self,
            other,
            schema=schema,
            exclude_schema=exclude_schema,
            ignore_extension_versions=ignore_extension_versions,
        )
        m.set_safety(False)

        if create_extensions_only:
            m.add_extension_changes(drops=False)
        else:
            m.add_all_changes(privileges=with_privileges)
        return m.statements

    def schemadiff_as_sql(self, other, **kwargs):
        statements = self.schemadiff_as_statements(other, **kwargs)
        return statements.sql

    def sync_schema_to_match(
        self,
        other,
        prompt=None,
        matching_message=None,
        tempdb=None,
        raise_on_cancel=True,
        **kwargs,
    ) -> str:
        from .database import Database
        from .tempdb import temporary_local_db

        if isinstance(other, str):
            tempdb = tempdb or temporary_local_db
            with tempdb() as other_db:
                other_db.q(other)
                statements = self.schemadiff_as_statements(other_db, **kwargs)

        elif isinstance(other, Database):
            other_db = other
            statements = self.schemadiff_as_statements(other_db, **kwargs)

        if not statements:
            if matching_message is not False:
                matching_text = matching_message or MATCHING_MESSAGE
                print(matching_text)

        statements_sql = statements.sql

        if prompt is not False:
            print(statements_sql)

            prompt_text = prompt or PROMPT_DEFAULT
            answer = input(prompt_text)

            if answer != "yes":
                print("Not syncing.")
                if raise_on_cancel:
                    raise ValueError('Sync cancelled due to non-"yes" prompt answer.')
                else:
                    return False

        self.q(statements_sql, fail_on_empty=False)
        return statements_sql



import random
import time
from contextlib import contextmanager
from string import ascii_lowercase

import results


def random_name(prefix):
    """Generate a random name with prefix, timestamp, and random suffix."""
    time_str = int(time.time())
    random_str = "".join([random.choice(ascii_lowercase) for _ in range(6)])

    return f"{prefix}-{time_str}-{random_str}"


@contextmanager
def temporary_local_db(url=None):
    """Create a temporary local database that is automatically cleaned up."""
    db_name = random_name("tempdb-pg").replace("-", "_")

    home_db = results.db(url)
    temp_db = home_db.create_db(db_name)

    try:
        yield temp_db
    finally:
        # Close the connection pool for this database before dropping it
        # This prevents "server closed the connection unexpectedly" errors
        from results.database import close_pool

        close_pool(temp_db.url)
        home_db.drop_db(db_name, yes_really_drop=True, force=True)



type_codes = {
    16: "bool",
    17: "bytea",
    18: "char",
    19: "name",
    20: "int8",
    21: "int2",
    22: "int2vector",
    23: "int4",
    24: "regproc",
    25: "text",
    26: "oid",
    27: "tid",
    28: "xid",
    29: "cid",
    30: "oidvector",
    71: "pg_type",
    75: "pg_attribute",
    81: "pg_proc",
    83: "pg_class",
    114: "json",
    142: "xml",
    194: "pg_node_tree",
    3361: "pg_ndistinct",
    3402: "pg_dependencies",
    5017: "pg_mcv_list",
    32: "pg_ddl_command",
    5069: "xid8",
    600: "point",
    601: "lseg",
    602: "path",
    603: "box",
    604: "polygon",
    628: "line",
    700: "float4",
    701: "float8",
    705: "unknown",
    718: "circle",
    790: "money",
    829: "macaddr",
    869: "inet",
    650: "cidr",
    774: "macaddr8",
    1033: "aclitem",
    1042: "bpchar",
    1043: "varchar",
    1082: "date",
    1083: "time",
    1114: "timestamp",
    1184: "timestamptz",
    1186: "interval",
    1266: "timetz",
    1560: "bit",
    1562: "varbit",
    1700: "numeric",
    1790: "refcursor",
    2202: "regprocedure",
    2203: "regoper",
    2204: "regoperator",
    2205: "regclass",
    4191: "regcollation",
    2206: "regtype",
    4096: "regrole",
    4089: "regnamespace",
    2950: "uuid",
    3220: "pg_lsn",
    3614: "tsvector",
    3642: "gtsvector",
    3615: "tsquery",
    3734: "regconfig",
    3769: "regdictionary",
    3802: "jsonb",
    4072: "jsonpath",
    2970: "txid_snapshot",
    5038: "pg_snapshot",
    3904: "int4range",
    3906: "numrange",
    3908: "tsrange",
    3910: "tstzrange",
    3912: "daterange",
    3926: "int8range",
    4451: "int4multirange",
    4532: "nummultirange",
    4533: "tsmultirange",
    4534: "tstzmultirange",
    4535: "datemultirange",
    4536: "int8multirange",
    2249: "record",
    2287: "_record",
    2275: "cstring",
    2276: "any",
    2277: "anyarray",
    2278: "void",
    2279: "trigger",
    3838: "event_trigger",
    2280: "language_handler",
    2281: "internal",
    2283: "anyelement",
    2776: "anynonarray",
    3500: "anyenum",
    3115: "fdw_handler",
    325: "index_am_handler",
    3310: "tsm_handler",
    269: "table_am_handler",
    3831: "anyrange",
    5077: "anycompatible",
    5078: "anycompatiblearray",
    5079: "anycompatiblenonarray",
    5080: "anycompatiblerange",
    4537: "anymultirange",
    4538: "anycompatiblemultirange",
    4600: "pg_brin_bloom_summary",
    4601: "pg_brin_minmax_multi_summary",
    1000: "_bool",
    1001: "_bytea",
    1002: "_char",
    1003: "_name",
    1016: "_int8",
    1005: "_int2",
    1006: "_int2vector",
    1007: "_int4",
    1008: "_regproc",
    1009: "_text",
    1028: "_oid",
    1010: "_tid",
    1011: "_xid",
    1012: "_cid",
    1013: "_oidvector",
    210: "_pg_type",
    270: "_pg_attribute",
    272: "_pg_proc",
    273: "_pg_class",
    199: "_json",
    143: "_xml",
    271: "_xid8",
    1017: "_point",
    1018: "_lseg",
    1019: "_path",
    1020: "_box",
    1027: "_polygon",
    629: "_line",
    1021: "_float4",
    1022: "_float8",
    719: "_circle",
    791: "_money",
    1040: "_macaddr",
    1041: "_inet",
    651: "_cidr",
    775: "_macaddr8",
    1034: "_aclitem",
    1014: "_bpchar",
    1015: "_varchar",
    1182: "_date",
    1183: "_time",
    1115: "_timestamp",
    1185: "_timestamptz",
    1187: "_interval",
    1270: "_timetz",
    1561: "_bit",
    1563: "_varbit",
    1231: "_numeric",
    2201: "_refcursor",
    2207: "_regprocedure",
    2208: "_regoper",
    2209: "_regoperator",
    2210: "_regclass",
    4192: "_regcollation",
    2211: "_regtype",
    4097: "_regrole",
    4090: "_regnamespace",
    2951: "_uuid",
    3221: "_pg_lsn",
    3643: "_tsvector",
    3644: "_gtsvector",
    3645: "_tsquery",
    3735: "_regconfig",
    3770: "_regdictionary",
    3807: "_jsonb",
    4073: "_jsonpath",
    2949: "_txid_snapshot",
    5039: "_pg_snapshot",
    3905: "_int4range",
    3907: "_numrange",
    3909: "_tsrange",
    3911: "_tstzrange",
    3913: "_daterange",
    3927: "_int8range",
    6150: "_int4multirange",
    6151: "_nummultirange",
    6152: "_tsmultirange",
    6153: "_tstzmultirange",
    6155: "_datemultirange",
    6157: "_int8multirange",
    1263: "_cstring",
    12001: "pg_attrdef",
    12000: "_pg_attrdef",
    12003: "pg_constraint",
    12002: "_pg_constraint",
    12005: "pg_inherits",
    12004: "_pg_inherits",
    12007: "pg_index",
    12006: "_pg_index",
    12009: "pg_operator",
    12008: "_pg_operator",
    12011: "pg_opfamily",
    12010: "_pg_opfamily",
    12013: "pg_opclass",
    12012: "_pg_opclass",
    12015: "pg_am",
    12014: "_pg_am",
    12017: "pg_amop",
    12016: "_pg_amop",
    12019: "pg_amproc",
    12018: "_pg_amproc",
    12021: "pg_language",
    12020: "_pg_language",
    12023: "pg_largeobject_metadata",
    12022: "_pg_largeobject_metadata",
    12025: "pg_largeobject",
    12024: "_pg_largeobject",
    12027: "pg_aggregate",
    12026: "_pg_aggregate",
    12029: "pg_statistic",
    12028: "_pg_statistic",
    12031: "pg_statistic_ext",
    12030: "_pg_statistic_ext",
    12033: "pg_statistic_ext_data",
    12032: "_pg_statistic_ext_data",
    12035: "pg_rewrite",
    12034: "_pg_rewrite",
    12037: "pg_trigger",
    12036: "_pg_trigger",
    12039: "pg_event_trigger",
    12038: "_pg_event_trigger",
    12041: "pg_description",
    12040: "_pg_description",
    12043: "pg_cast",
    12042: "_pg_cast",
    12045: "pg_enum",
    12044: "_pg_enum",
    12047: "pg_namespace",
    12046: "_pg_namespace",
    12049: "pg_conversion",
    12048: "_pg_conversion",
    12051: "pg_depend",
    12050: "_pg_depend",
    1248: "pg_database",
    12052: "_pg_database",
    12054: "pg_db_role_setting",
    12053: "_pg_db_role_setting",
    12056: "pg_tablespace",
    12055: "_pg_tablespace",
    2842: "pg_authid",
    12057: "_pg_authid",
    2843: "pg_auth_members",
    12058: "_pg_auth_members",
    12060: "pg_shdepend",
    12059: "_pg_shdepend",
    12062: "pg_shdescription",
    12061: "_pg_shdescription",
    12064: "pg_ts_config",
    12063: "_pg_ts_config",
    12066: "pg_ts_config_map",
    12065: "_pg_ts_config_map",
    12068: "pg_ts_dict",
    12067: "_pg_ts_dict",
    12070: "pg_ts_parser",
    12069: "_pg_ts_parser",
    12072: "pg_ts_template",
    12071: "_pg_ts_template",
    12074: "pg_extension",
    12073: "_pg_extension",
    12076: "pg_foreign_data_wrapper",
    12075: "_pg_foreign_data_wrapper",
    12078: "pg_foreign_server",
    12077: "_pg_foreign_server",
    12080: "pg_user_mapping",
    12079: "_pg_user_mapping",
    12082: "pg_foreign_table",
    12081: "_pg_foreign_table",
    12084: "pg_policy",
    12083: "_pg_policy",
    12086: "pg_replication_origin",
    12085: "_pg_replication_origin",
    12088: "pg_default_acl",
    12087: "_pg_default_acl",
    12090: "pg_init_privs",
    12089: "_pg_init_privs",
    12092: "pg_seclabel",
    12091: "_pg_seclabel",
    4066: "pg_shseclabel",
    12093: "_pg_shseclabel",
    12095: "pg_collation",
    12094: "_pg_collation",
    12097: "pg_partitioned_table",
    12096: "_pg_partitioned_table",
    12099: "pg_range",
    12098: "_pg_range",
    12101: "pg_transform",
    12100: "_pg_transform",
    12103: "pg_sequence",
    12102: "_pg_sequence",
    12105: "pg_publication",
    12104: "_pg_publication",
    12107: "pg_publication_rel",
    12106: "_pg_publication_rel",
    6101: "pg_subscription",
    12108: "_pg_subscription",
    12110: "pg_subscription_rel",
    12109: "_pg_subscription_rel",
    12219: "pg_roles",
    12218: "_pg_roles",
    12224: "pg_shadow",
    12223: "_pg_shadow",
    12229: "pg_group",
    12228: "_pg_group",
    12233: "pg_user",
    12232: "_pg_user",
    12237: "pg_policies",
    12236: "_pg_policies",
    12242: "pg_rules",
    12241: "_pg_rules",
    12247: "pg_views",
    12246: "_pg_views",
    12252: "pg_tables",
    12251: "_pg_tables",
    12257: "pg_matviews",
    12256: "_pg_matviews",
    12262: "pg_indexes",
    12261: "_pg_indexes",
    12267: "pg_sequences",
    12266: "_pg_sequences",
    12272: "pg_stats",
    12271: "_pg_stats",
    12277: "pg_stats_ext",
    12276: "_pg_stats_ext",
    12282: "pg_stats_ext_exprs",
    12281: "_pg_stats_ext_exprs",
    12287: "pg_publication_tables",
    12286: "_pg_publication_tables",
    12292: "pg_locks",
    12291: "_pg_locks",
    12296: "pg_cursors",
    12295: "_pg_cursors",
    12300: "pg_available_extensions",
    12299: "_pg_available_extensions",
    12304: "pg_available_extension_versions",
    12303: "_pg_available_extension_versions",
    12309: "pg_prepared_xacts",
    12308: "_pg_prepared_xacts",
    12314: "pg_prepared_statements",
    12313: "_pg_prepared_statements",
    12318: "pg_seclabels",
    12317: "_pg_seclabels",
    12323: "pg_settings",
    12322: "_pg_settings",
    12329: "pg_file_settings",
    12328: "_pg_file_settings",
    12333: "pg_hba_file_rules",
    12332: "_pg_hba_file_rules",
    12337: "pg_timezone_abbrevs",
    12336: "_pg_timezone_abbrevs",
    12341: "pg_timezone_names",
    12340: "_pg_timezone_names",
    12345: "pg_config",
    12344: "_pg_config",
    12349: "pg_shmem_allocations",
    12348: "_pg_shmem_allocations",
    12353: "pg_backend_memory_contexts",
    12352: "_pg_backend_memory_contexts",
    12357: "pg_stat_all_tables",
    12356: "_pg_stat_all_tables",
    12362: "pg_stat_xact_all_tables",
    12361: "_pg_stat_xact_all_tables",
    12367: "pg_stat_sys_tables",
    12366: "_pg_stat_sys_tables",
    12372: "pg_stat_xact_sys_tables",
    12371: "_pg_stat_xact_sys_tables",
    12376: "pg_stat_user_tables",
    12375: "_pg_stat_user_tables",
    12381: "pg_stat_xact_user_tables",
    12380: "_pg_stat_xact_user_tables",
    12385: "pg_statio_all_tables",
    12384: "_pg_statio_all_tables",
    12390: "pg_statio_sys_tables",
    12389: "_pg_statio_sys_tables",
    12394: "pg_statio_user_tables",
    12393: "_pg_statio_user_tables",
    12398: "pg_stat_all_indexes",
    12397: "_pg_stat_all_indexes",
    12403: "pg_stat_sys_indexes",
    12402: "_pg_stat_sys_indexes",
    12407: "pg_stat_user_indexes",
    12406: "_pg_stat_user_indexes",
    12411: "pg_statio_all_indexes",
    12410: "_pg_statio_all_indexes",
    12416: "pg_statio_sys_indexes",
    12415: "_pg_statio_sys_indexes",
    12420: "pg_statio_user_indexes",
    12419: "_pg_statio_user_indexes",
    12424: "pg_statio_all_sequences",
    12423: "_pg_statio_all_sequences",
    12429: "pg_statio_sys_sequences",
    12428: "_pg_statio_sys_sequences",
    12433: "pg_statio_user_sequences",
    12432: "_pg_statio_user_sequences",
    12437: "pg_stat_activity",
    12436: "_pg_stat_activity",
    12442: "pg_stat_replication",
    12441: "_pg_stat_replication",
    12447: "pg_stat_slru",
    12446: "_pg_stat_slru",
    12451: "pg_stat_wal_receiver",
    12450: "_pg_stat_wal_receiver",
    12455: "pg_stat_subscription",
    12454: "_pg_stat_subscription",
    12460: "pg_stat_ssl",
    12459: "_pg_stat_ssl",
    12464: "pg_stat_gssapi",
    12463: "_pg_stat_gssapi",
    12468: "pg_replication_slots",
    12467: "_pg_replication_slots",
    12473: "pg_stat_replication_slots",
    12472: "_pg_stat_replication_slots",
    12477: "pg_stat_database",
    12476: "_pg_stat_database",
    12482: "pg_stat_database_conflicts",
    12481: "_pg_stat_database_conflicts",
    12486: "pg_stat_user_functions",
    12485: "_pg_stat_user_functions",
    12491: "pg_stat_xact_user_functions",
    12490: "_pg_stat_xact_user_functions",
    12496: "pg_stat_archiver",
    12495: "_pg_stat_archiver",
    12500: "pg_stat_bgwriter",
    12499: "_pg_stat_bgwriter",
    12504: "pg_stat_wal",
    12503: "_pg_stat_wal",
    12508: "pg_stat_progress_analyze",
    12507: "_pg_stat_progress_analyze",
    12513: "pg_stat_progress_vacuum",
    12512: "_pg_stat_progress_vacuum",
    12518: "pg_stat_progress_cluster",
    12517: "_pg_stat_progress_cluster",
    12523: "pg_stat_progress_create_index",
    12522: "_pg_stat_progress_create_index",
    12528: "pg_stat_progress_basebackup",
    12527: "_pg_stat_progress_basebackup",
    12533: "pg_stat_progress_copy",
    12532: "_pg_stat_progress_copy",
    12538: "pg_user_mappings",
    12537: "_pg_user_mappings",
    12543: "pg_replication_origin_status",
    12542: "_pg_replication_origin_status",
    13669: "cardinal_number",
    13668: "_cardinal_number",
    13672: "character_data",
    13671: "_character_data",
    13674: "sql_identifier",
    13673: "_sql_identifier",
    13677: "information_schema_catalog_name",
    13676: "_information_schema_catalog_name",
    13680: "time_stamp",
    13679: "_time_stamp",
    13682: "yes_or_no",
    13681: "_yes_or_no",
    13686: "applicable_roles",
    13685: "_applicable_roles",
    13691: "administrable_role_authorizations",
    13690: "_administrable_role_authorizations",
    13695: "attributes",
    13694: "_attributes",
    13700: "character_sets",
    13699: "_character_sets",
    13705: "check_constraint_routine_usage",
    13704: "_check_constraint_routine_usage",
    13710: "check_constraints",
    13709: "_check_constraints",
    13715: "collations",
    13714: "_collations",
    13720: "collation_character_set_applicability",
    13719: "_collation_character_set_applicability",
    13725: "column_column_usage",
    13724: "_column_column_usage",
    13730: "column_domain_usage",
    13729: "_column_domain_usage",
    13735: "column_privileges",
    13734: "_column_privileges",
    13740: "column_udt_usage",
    13739: "_column_udt_usage",
    13745: "columns",
    13744: "_columns",
    13750: "constraint_column_usage",
    13749: "_constraint_column_usage",
    13755: "constraint_table_usage",
    13754: "_constraint_table_usage",
    13760: "domain_constraints",
    13759: "_domain_constraints",
    13765: "domain_udt_usage",
    13764: "_domain_udt_usage",
    13770: "domains",
    13769: "_domains",
    13775: "enabled_roles",
    13774: "_enabled_roles",
    13779: "key_column_usage",
    13778: "_key_column_usage",
    13784: "parameters",
    13783: "_parameters",
    13789: "referential_constraints",
    13788: "_referential_constraints",
    13794: "role_column_grants",
    13793: "_role_column_grants",
    13798: "routine_column_usage",
    13797: "_routine_column_usage",
    13803: "routine_privileges",
    13802: "_routine_privileges",
    13808: "role_routine_grants",
    13807: "_role_routine_grants",
    13812: "routine_routine_usage",
    13811: "_routine_routine_usage",
    13817: "routine_sequence_usage",
    13816: "_routine_sequence_usage",
    13822: "routine_table_usage",
    13821: "_routine_table_usage",
    13827: "routines",
    13826: "_routines",
    13832: "schemata",
    13831: "_schemata",
    13836: "sequences",
    13835: "_sequences",
    13841: "sql_features",
    13840: "_sql_features",
    13846: "sql_implementation_info",
    13845: "_sql_implementation_info",
    13851: "sql_parts",
    13850: "_sql_parts",
    13856: "sql_sizing",
    13855: "_sql_sizing",
    13861: "table_constraints",
    13860: "_table_constraints",
    13866: "table_privileges",
    13865: "_table_privileges",
    13871: "role_table_grants",
    13870: "_role_table_grants",
    13875: "tables",
    13874: "_tables",
    13880: "transforms",
    13879: "_transforms",
    13885: "triggered_update_columns",
    13884: "_triggered_update_columns",
    13890: "triggers",
    13889: "_triggers",
    13895: "udt_privileges",
    13894: "_udt_privileges",
    13900: "role_udt_grants",
    13899: "_role_udt_grants",
    13904: "usage_privileges",
    13903: "_usage_privileges",
    13909: "role_usage_grants",
    13908: "_role_usage_grants",
    13913: "user_defined_types",
    13912: "_user_defined_types",
    13918: "view_column_usage",
    13917: "_view_column_usage",
    13923: "view_routine_usage",
    13922: "_view_routine_usage",
    13928: "view_table_usage",
    13927: "_view_table_usage",
    13933: "views",
    13932: "_views",
    13938: "data_type_privileges",
    13937: "_data_type_privileges",
    13943: "element_types",
    13942: "_element_types",
    13948: "_pg_foreign_table_columns",
    13947: "__pg_foreign_table_columns",
    13953: "column_options",
    13952: "_column_options",
    13957: "_pg_foreign_data_wrappers",
    13956: "__pg_foreign_data_wrappers",
    13961: "foreign_data_wrapper_options",
    13960: "_foreign_data_wrapper_options",
    13965: "foreign_data_wrappers",
    13964: "_foreign_data_wrappers",
    13969: "_pg_foreign_servers",
    13968: "__pg_foreign_servers",
    13974: "foreign_server_options",
    13973: "_foreign_server_options",
    13978: "foreign_servers",
    13977: "_foreign_servers",
    13982: "_pg_foreign_tables",
    13981: "__pg_foreign_tables",
    13987: "foreign_table_options",
    13986: "_foreign_table_options",
    13991: "foreign_tables",
    13990: "_foreign_tables",
    13995: "_pg_user_mappings",
    13994: "__pg_user_mappings",
    14000: "user_mapping_options",
    13999: "_user_mapping_options",
    14005: "user_mappings",
    14004: "_user_mappings",
    1261458: "citext",
    1261463: "_citext",
    4382494: "circles",
    4382493: "_circles",
}



from results.schemainspect import PostgreSQL, get_inspector

from .changes import Changes
from .statements import Statements


class Migration:
    """
    The main class of migra
    """

    def __init__(
        self,
        x_from,
        x_target,
        schema=None,
        exclude_schema=None,
        ignore_extension_versions=False,
    ):
        self.statements = Statements()
        self.changes = Changes(None, None)
        if schema and exclude_schema:
            raise ValueError("You cannot have both a schema and excluded schema")
        self.schema = schema
        self.exclude_schema = exclude_schema
        if isinstance(x_from, PostgreSQL):
            self.changes.i_from = x_from
        else:
            self.changes.i_from = get_inspector(
                x_from, schema=schema, exclude_schema=exclude_schema
            )
            if x_from:
                self.s_from = x_from
        if isinstance(x_target, PostgreSQL):
            self.changes.i_target = x_target
        else:
            self.changes.i_target = get_inspector(
                x_target, schema=schema, exclude_schema=exclude_schema
            )
            if x_target:
                self.s_target = x_target

        self.changes.ignore_extension_versions = ignore_extension_versions

    def inspect_from(self):
        self.changes.i_from = get_inspector(
            self.s_from, schema=self.schema, exclude_schema=self.exclude_schema
        )

    def inspect_target(self):
        self.changes.i_target = get_inspector(
            self.s_target, schema=self.schema, exclude_schema=self.exclude_schema
        )

    def clear(self):
        self.statements = Statements()

    def add(self, statements):
        self.statements += statements

    def add_sql(self, sql):
        self.statements += Statements([sql])

    def set_safety(self, safety_on):
        self.statements.safe = safety_on

    def add_extension_changes(self, creates=True, drops=True):
        if creates:
            self.add(self.changes.extensions(creations_only=True))
        if drops:
            self.add(self.changes.extensions(drops_only=True))

    def add_all_changes(self, privileges=False):
        self.add(self.changes.schemas(creations_only=True))

        self.add(self.changes.extensions(creations_only=True, modifications=False))
        self.add(self.changes.extensions(modifications_only=True, modifications=True))
        self.add(self.changes.collations(creations_only=True))
        self.add(self.changes.enums(creations_only=True, modifications=False))
        self.add(self.changes.types(creations_only=True))
        self.add(self.changes.types(modifications_only=True))
        self.add(self.changes.sequences(creations_only=True))
        self.add(self.changes.triggers(drops_only=True))
        self.add(self.changes.rlspolicies(drops_only=True))
        if privileges:
            self.add(self.changes.privileges(drops_only=True))
        self.add(self.changes.non_pk_constraints(drops_only=True))

        self.add(self.changes.mv_indexes(drops_only=True))
        self.add(self.changes.non_table_selectable_drops())

        self.add(self.changes.pk_constraints(drops_only=True))
        self.add(self.changes.non_mv_indexes(drops_only=True))

        self.add(self.changes.tables_only_selectables())

        self.add(self.changes.sequences(drops_only=True))
        self.add(self.changes.enums(drops_only=True, modifications=False))
        self.add(self.changes.types(drops_only=True))
        self.add(self.changes.extensions(drops_only=True, modifications=False))
        self.add(self.changes.non_mv_indexes(creations_only=True))
        self.add(self.changes.pk_constraints(creations_only=True))
        self.add(self.changes.non_pk_constraints(creations_only=True))

        self.add(self.changes.non_table_selectable_creations())
        self.add(self.changes.mv_indexes(creations_only=True))

        if privileges:
            self.add(self.changes.privileges(creations_only=True))
        self.add(self.changes.rlspolicies(creations_only=True))
        self.add(self.changes.triggers(creations_only=True))
        self.add(self.changes.collations(drops_only=True))
        self.add(self.changes.schemas(drops_only=True))

    @property
    def sql(self):
        return self.statements.sql



import csv
from io import StringIO

from .rows import Rows


def rows_from_text(text, format="csv", **kwargs):
    """Parse text data and return a Rows object."""
    reader = csv.reader(StringIO(text.strip()))

    column_names = next(reader)

    rows = Rows(row for row in reader)
    rows.column_info = {name: "text" for name in column_names}

    return rows



import textwrap
from collections import OrderedDict as od
from itertools import groupby

from ..inspected import ColumnInfo, Inspected
from ..inspected import InspectedSelectable as BaseInspectedSelectable
from ..inspected import TableRelated
from ..misc import quoted_identifier, resource_text

CREATE_TABLE = """create {}table {} ({}
){}{};
"""
CREATE_TABLE_SUBCLASS = """create {}table {} partition of {} {};
"""
CREATE_FUNCTION_FORMAT = """create or replace function {signature}
returns {result_string} as
$${definition}$$
language {language} {volatility} {strictness} {security_type};"""
ALL_RELATIONS_QUERY = resource_text("sql/relations.sql")
SCHEMAS_QUERY = resource_text("sql/schemas.sql")
INDEXES_QUERY = resource_text("sql/indexes.sql")
SEQUENCES_QUERY = resource_text("sql/sequences.sql")
CONSTRAINTS_QUERY = resource_text("sql/constraints.sql")
FUNCTIONS_QUERY = resource_text("sql/functions.sql")
TYPES_QUERY = resource_text("sql/types.sql")
DOMAINS_QUERY = resource_text("sql/domains.sql")
EXTENSIONS_QUERY = resource_text("sql/extensions.sql")
ENUMS_QUERY = resource_text("sql/enums.sql")
DEPS_QUERY = resource_text("sql/deps.sql")
PRIVILEGES_QUERY = resource_text("sql/privileges.sql")
TRIGGERS_QUERY = resource_text("sql/triggers.sql")
COLLATIONS_QUERY = resource_text("sql/collations.sql")
RLSPOLICIES_QUERY = resource_text("sql/rlspolicies.sql")
EXTOIDS_QUERY = resource_text("sql/extoids.sql")


class AttDict(dict):
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)


class InspectedSelectable(BaseInspectedSelectable):
    def has_compatible_columns(self, other):
        def names_and_types(cols):
            return [(k, c.dbtype) for k, c in cols.items()]

        items = names_and_types(self.columns)

        if self.relationtype != "f":
            old_arg_count = len(other.columns)
            items = items[:old_arg_count]

        return items == names_and_types(other.columns)

    def can_replace(self, other):
        if not (self.relationtype in ("v", "f") or self.is_table):
            return False

        if self.signature != other.signature:
            return False

        if self.relationtype != other.relationtype:
            return False

        return self.has_compatible_columns(other)

    @property
    def persistence_modifier(self):
        if self.persistence == "t":
            return "temporary "
        elif self.persistence == "u":
            return "unlogged "
        else:
            return ""

    @property
    def is_unlogged(self):
        return self.persistence == "u"

    @property
    def create_statement(self):
        n = self.quoted_full_name
        if self.relationtype in ("r", "p"):
            if not self.is_partitioning_child_table:
                colspec = ",\n".join(
                    "    " + c.creation_clause for c in self.columns.values()
                )
                if colspec:
                    colspec = "\n" + colspec

                if self.partition_def:
                    partition_key = " partition by " + self.partition_def
                    inherits_clause = ""
                elif self.parent_table:
                    inherits_clause = " inherits ({})".format(self.parent_table)
                    partition_key = ""
                else:
                    partition_key = ""
                    inherits_clause = ""

                create_statement = CREATE_TABLE.format(
                    self.persistence_modifier,
                    n,
                    colspec,
                    partition_key,
                    inherits_clause,
                )
            else:
                create_statement = CREATE_TABLE_SUBCLASS.format(
                    self.persistence_modifier, n, self.parent_table, self.partition_def
                )
        elif self.relationtype == "v":
            with_options = (
                " with ({})".format(", ".join(self.options)) if self.options else ""
            )
            create_statement = "create or replace view {}{} as {}\n".format(
                n, with_options, self.definition
            )
        elif self.relationtype == "m":
            create_statement = "create materialized view {} as {}\n".format(
                n, self.definition
            )
        elif self.relationtype == "c":
            colspec = ", ".join(c.creation_clause for c in self.columns.values())
            create_statement = "create type {} as ({});".format(n, colspec)
        else:
            raise NotImplementedError  # pragma: no cover
        return create_statement

    @property
    def drop_statement(self):
        n = self.quoted_full_name
        if self.relationtype in ("r", "p"):
            drop_statement = "drop table {};".format(n)
        elif self.relationtype == "v":
            drop_statement = "drop view if exists {};".format(n)
        elif self.relationtype == "m":
            drop_statement = "drop materialized view if exists {};".format(n)
        elif self.relationtype == "c":
            drop_statement = "drop type {};".format(n)
        else:
            raise NotImplementedError  # pragma: no cover

        return drop_statement

    def alter_table_statement(self, clause):
        if self.is_alterable:
            alter = "alter table {} {};".format(self.quoted_full_name, clause)
        else:
            raise NotImplementedError  # pragma: no cover

        return alter

    @property
    def is_partitioned(self):
        return self.relationtype == "p"

    @property
    def is_inheritance_child_table(self):
        return bool(self.parent_table) and not self.partition_def

    @property
    def is_table(self):
        return self.relationtype in ("p", "r")

    @property
    def is_alterable(self):
        return self.is_table and (
            not self.parent_table or self.is_inheritance_child_table
        )

    @property
    def contains_data(self):
        return bool(
            self.relationtype == "r" and (self.parent_table or not self.partition_def)
        )

    # for back-compat only
    @property
    def is_child_table(self):
        return self.is_partitioning_child_table

    @property
    def is_partitioning_child_table(self):
        return bool(
            self.relationtype == "r" and self.parent_table and self.partition_def
        )

    @property
    def uses_partitioning(self):
        return self.is_partitioning_child_table or self.is_partitioned

    @property
    def attach_statement(self):
        if self.parent_table:
            if self.partition_def:
                return "alter table {} attach partition {} {};".format(
                    self.quoted_full_name, self.parent_table, self.partition_spec
                )
            else:
                return "alter table {} inherit {}".format(
                    self.quoted_full_name, self.parent_table
                )

    @property
    def detach_statement(self):
        if self.parent_table:
            if self.partition_def:
                return "alter table {} detach partition {};".format(
                    self.parent_table, self.quoted_full_name
                )
            else:
                return "alter table {} no inherit {}".format(
                    self.quoted_full_name, self.parent_table
                )

    def attach_detach_statements(self, before):
        slist = []
        if self.parent_table != before.parent_table:
            if before.parent_table:
                slist.append(before.detach_statement)
            if self.parent_table:
                slist.append(self.attach_statement)
        return slist

    @property
    def alter_rls_clause(self):
        keyword = "enable" if self.rowsecurity else "disable"
        return "{} row level security".format(keyword)

    @property
    def alter_rls_statement(self):
        return self.alter_table_statement(self.alter_rls_clause)

    @property
    def alter_unlogged_statement(self):
        keyword = "unlogged" if self.is_unlogged else "logged"
        return self.alter_table_statement("set {}".format(keyword))

    def alter_statements(self, other):
        """Generate ALTER statements to transform the other object to match this one."""
        return other.comment_alter_statements(self)


class InspectedFunction(InspectedSelectable):
    def __init__(
        self,
        name,
        schema,
        columns,
        inputs,
        definition,
        volatility,
        strictness,
        security_type,
        identity_arguments,
        result_string,
        language,
        full_definition,
        comment,
        returntype,
        kind,
    ):
        self.identity_arguments = identity_arguments
        self.result_string = result_string
        self.language = language
        self.volatility = volatility
        self.strictness = strictness
        self.security_type = security_type
        self.full_definition = full_definition
        self.returntype = returntype
        self.kind = kind

        super(InspectedFunction, self).__init__(
            name=name,
            schema=schema,
            columns=columns,
            inputs=inputs,
            definition=definition,
            relationtype="f",
            comment=comment,
        )

    @property
    def returntype_is_table(self):
        if self.returntype:
            return "." in self.returntype

    @property
    def signature(self):
        return "{}({})".format(self.quoted_full_name, self.identity_arguments)

    @property
    def create_statement(self):
        return self.full_definition + ";"
        """
        return CREATE_FUNCTION_FORMAT.format(
            signature=self.signature,
            result_string=self.result_string,
            definition=self.definition,
            language=self.language,
            volatility=self.volatility,
            strictness=self.strictness,
            security_type=self.security_type,
        )
        """

    @property
    def thing(self):
        kinds = dict(f="function", p="procedure", a="aggregate", w="window function")
        return kinds[self.kind]

    @property
    def drop_statement(self):
        return "drop {} if exists {};".format(self.thing, self.signature)

    def __eq__(self, other):
        return (
            self.signature == other.signature
            and self.result_string == other.result_string
            and self.definition == other.definition
            and self.language == other.language
            and self.volatility == other.volatility
            and self.strictness == other.strictness
            and self.security_type == other.security_type
            and self.kind == other.kind
            and self.comment == other.comment
        )

    @property
    def comment_statement(self):
        """Generate a COMMENT ON statement for this function."""
        if self.comment is None:
            return None

        escaped_comment = self.comment.replace("'", "''")
        return f"COMMENT ON FUNCTION {self.signature} IS '{escaped_comment}';"

    @property
    def drop_comment_statement(self):
        """Generate a statement to drop the comment for this function."""
        return f"COMMENT ON FUNCTION {self.signature} IS NULL;"

    def comment_alter_statements(self, other):
        """Generate statements to alter comments between two functions."""
        statements = []

        if self.comment != other.comment:
            if other.comment:
                # Adding or changing a comment
                escaped_comment = other.comment.replace("'", "''")
                statements.append(
                    f"COMMENT ON FUNCTION {other.signature} IS '{escaped_comment}';"
                )
            else:
                # Removing a comment
                statements.append(self.drop_comment_statement)

        return statements


class InspectedTrigger(Inspected):
    def __init__(
        self, name, schema, table_name, proc_schema, proc_name, enabled, full_definition
    ):
        (
            self.name,
            self.schema,
            self.table_name,
            self.proc_schema,
            self.proc_name,
            self.enabled,
            self.full_definition,
        ) = (name, schema, table_name, proc_schema, proc_name, enabled, full_definition)

        self.dependent_on = [self.quoted_full_selectable_name]
        self.dependents = []

    @property
    def signature(self):
        return self.quoted_full_name

    @property
    def quoted_full_name(self):
        return "{}.{}.{}".format(
            quoted_identifier(self.schema),
            quoted_identifier(self.table_name),
            quoted_identifier(self.name),
        )

    @property
    def quoted_full_selectable_name(self):
        return "{}.{}".format(
            quoted_identifier(self.schema), quoted_identifier(self.table_name)
        )

    @property
    def drop_statement(self):
        return 'drop trigger if exists "{}" on "{}"."{}";'.format(
            self.name, self.schema, self.table_name
        )

    @property
    def create_statement(self):
        status_sql = {
            "O": "ENABLE TRIGGER",
            "D": "DISABLE TRIGGER",
            "R": "ENABLE REPLICA TRIGGER",
            "A": "ENABLE ALWAYS TRIGGER",
        }
        schema = quoted_identifier(self.schema)
        table = quoted_identifier(self.table_name)
        trigger_name = quoted_identifier(self.name)
        if self.enabled in ("D", "R", "A"):
            table_alter = f"ALTER TABLE {schema}.{table} {status_sql[self.enabled]} {trigger_name}"
            return self.full_definition + ";\n" + table_alter + ";"
        else:
            return self.full_definition + ";"

    def __eq__(self, other):
        """
        :type other: InspectedTrigger
        :rtype: bool
        """
        return (
            self.name == other.name
            and self.schema == other.schema
            and self.table_name == other.table_name
            and self.proc_schema == other.proc_schema
            and self.proc_name == other.proc_name
            and self.enabled == other.enabled
            and self.full_definition == other.full_definition
        )


class InspectedIndex(Inspected, TableRelated):
    def __init__(
        self,
        name,
        schema,
        table_name,
        key_columns,
        key_options,
        num_att,
        is_unique,
        is_pk,
        is_exclusion,
        is_immediate,
        is_clustered,
        key_collations,
        key_expressions,
        partial_predicate,
        algorithm,
        definition=None,
        constraint=None,
        index_columns=None,
        included_columns=None,
    ):
        self.name = name
        self.schema = schema
        self.definition = definition
        self.table_name = table_name
        self.key_columns = key_columns
        self.key_options = key_options
        self.num_att = num_att
        self.is_unique = is_unique
        self.is_pk = is_pk
        self.is_exclusion = is_exclusion
        self.is_immediate = is_immediate
        self.is_clustered = is_clustered
        self.key_collations = key_collations
        self.key_expressions = key_expressions
        self.partial_predicate = partial_predicate
        self.algorithm = algorithm
        self.constraint = constraint
        self.index_columns = index_columns
        self.included_columns = included_columns

    @property
    def drop_statement(self):
        statement = "drop index if exists {};".format(self.quoted_full_name)

        if self.is_exclusion_constraint:
            return "select 1; " + textwrap.indent(statement, "-- ")
        return statement

    @property
    def create_statement(self):
        statement = "{};".format(self.definition)
        if self.is_exclusion_constraint:
            return "select 1; " + textwrap.indent(statement, "-- ")
        return statement

    @property
    def is_exclusion_constraint(self):
        return self.constraint and self.constraint.constraint_type == "EXCLUDE"

    def __eq__(self, other):
        """
        :type other: InspectedIndex
        :rtype: bool
        """
        equalities = (
            self.name == other.name,
            self.schema == other.schema,
            self.table_name == other.table_name,
            self.key_columns == other.key_columns,
            self.included_columns == other.included_columns,
            self.key_options == other.key_options,
            self.num_att == other.num_att,
            self.is_unique == other.is_unique,
            self.is_pk == other.is_pk,
            self.is_exclusion == other.is_exclusion,
            self.is_immediate == other.is_immediate,
            self.is_clustered == other.is_clustered,
            self.key_expressions == other.key_expressions,
            self.partial_predicate == other.partial_predicate,
            self.algorithm == other.algorithm,
        )
        return all(equalities)


class InspectedSequence(Inspected):
    def __init__(self, name, schema, table_name=None, column_name=None):
        self.name = name
        self.schema = schema
        self.table_name = table_name
        self.column_name = column_name

    @property
    def drop_statement(self):
        return "drop sequence if exists {};".format(self.quoted_full_name)

    @property
    def create_statement(self):
        return "create sequence {};".format(self.quoted_full_name)

    @property
    def create_statement_with_ownership(self):
        t_col_name = self.quoted_table_and_column_name

        if self.table_name and self.column_name:
            return "create sequence {} owned by {};".format(
                self.quoted_full_name, t_col_name
            )
        else:
            return "create sequence {};".format(self.quoted_full_name)

    @property
    def alter_ownership_statement(self):
        t_col_name = self.quoted_table_and_column_name

        if t_col_name is not None:
            return "alter sequence {} owned by {};".format(
                self.quoted_full_name, t_col_name
            )
        else:
            return "alter sequence {} owned by none;".format(self.quoted_full_name)

    @property
    def quoted_full_table_name(self):
        if self.table_name is not None:
            return quoted_identifier(self.table_name, self.schema)

    @property
    def quoted_table_and_column_name(self):
        if self.column_name is not None and self.table_name is not None:
            return (
                self.quoted_full_table_name + "." + quoted_identifier(self.column_name)
            )

    def __eq__(self, other):
        equalities = (
            self.name == other.name,
            self.schema == other.schema,
            self.quoted_table_and_column_name == other.quoted_table_and_column_name,
        )
        return all(equalities)


class InspectedCollation(Inspected):
    def __init__(
        self,
        name,
        schema,
        provider,
        encoding,
        lc_collate,
        lc_ctype,
        version,
        locale,
    ):
        self.name = name
        self.schema = schema
        self.provider = provider
        self.lc_collate = lc_collate
        self.lc_ctype = lc_ctype
        self.encoding = encoding
        self.version = version
        self.locale = locale

    @property
    def drop_statement(self):
        return "drop collation if exists {};".format(self.quoted_full_name)

    @property
    def create_statement(self):
        return "create collation if not exists {} (provider = '{}', locale = '{}');".format(
            self.quoted_full_name, self.provider, self.locale
        )

    def __eq__(self, other):
        equalities = (
            self.name == other.name,
            self.schema == other.schema,
            self.provider == other.provider,
            self.locale == other.locale,
        )
        return all(equalities)


class InspectedEnum(Inspected):
    def __init__(self, name, schema, elements, pg_version=None):
        self.name = name
        self.schema = schema
        self.elements = elements
        self.pg_version = pg_version
        self.dependents = []
        self.dependent_on = []

    @property
    def drop_statement(self):
        return "drop type {};".format(self.quoted_full_name)

    @property
    def create_statement(self):
        return "create type {} as enum ({});".format(
            self.quoted_full_name, self.quoted_elements
        )

    @property
    def quoted_elements(self):
        quoted = ["'{}'".format(e) for e in self.elements]
        return ", ".join(quoted)

    def alter_rename_statement(self, new_name):
        name = new_name

        return "alter type {} rename to {};".format(
            self.quoted_full_name, quoted_identifier(name)
        )

    def drop_statement_with_rename(self, new_name):
        name = new_name
        new_name = quoted_identifier(name, self.schema)
        return "drop type {};".format(new_name)

    def change_statements(self, new):
        if not self.can_be_changed_to(new):
            raise ValueError

        new = new.elements
        old = self.elements
        statements = []
        previous = None
        for c in new:
            if c not in old:
                if not previous:
                    s = "alter type {} add value '{}' before '{}';".format(
                        self.quoted_full_name, c, old[0]
                    )
                else:
                    s = "alter type {} add value '{}' after '{}';".format(
                        self.quoted_full_name, c, previous
                    )
                statements.append(s)
            previous = c
        return statements

    def can_be_changed_to(self, new, when_within_transaction=False):
        old = self.elements

        if when_within_transaction and self.pg_version and self.pg_version < 12:
            return False

        # new must already have the existing items from old, in the same order
        return [e for e in new.elements if e in old] == old

    def __eq__(self, other):
        equalities = (
            self.name == other.name,
            self.schema == other.schema,
            self.elements == other.elements,
        )
        return all(equalities)


class InspectedSchema(Inspected):
    def __init__(self, schema):
        self.schema = schema
        self.name = None

    @property
    def create_statement(self):
        return "create schema if not exists {};".format(self.quoted_schema)

    @property
    def drop_statement(self):
        return "drop schema if exists {};".format(self.quoted_schema)

    @property
    def quoted_full_name(self):
        return self.quoted_name

    @property
    def quoted_name(self):
        return quoted_identifier(self.schema)

    def __eq__(self, other):
        return self.schema == other.schema


class InspectedType(Inspected):
    def __init__(self, name, schema, columns, comment=None):
        self.name = name
        self.schema = schema
        self.columns = columns
        self.comment = comment

    @property
    def drop_statement(self):
        return "drop type {};".format(self.signature)

    @property
    def create_statement(self):
        sql = "create type {} as (\n".format(self.signature)

        indent = " " * 4
        typespec = [
            "{}{} {}".format(indent, quoted_identifier(name), _type)
            for name, _type in self.columns.items()
        ]

        sql += ",\n".join(typespec)
        sql += "\n);"
        return sql

    def __eq__(self, other):
        return (
            self.schema == other.schema
            and self.name == other.name
            and self.columns == other.columns
            and self.comment == other.comment
        )

    @property
    def comment_statement(self):
        """Generate a COMMENT ON statement for this type."""
        if self.comment is None:
            return None

        escaped_comment = self.comment.replace("'", "''")
        return f"COMMENT ON TYPE {self.signature} IS '{escaped_comment}';"

    @property
    def drop_comment_statement(self):
        """Generate a statement to drop the comment for this type."""
        return f"COMMENT ON TYPE {self.signature} IS NULL;"

    def comment_alter_statements(self, other):
        """Generate statements to alter comments between two types."""
        statements = []

        if self.comment != other.comment:
            if other.comment:
                # Adding or changing a comment
                escaped_comment = other.comment.replace("'", "''")
                statements.append(
                    f"COMMENT ON TYPE {other.signature} IS '{escaped_comment}';"
                )
            else:
                # Removing a comment
                statements.append(self.drop_comment_statement)

        return statements

    def alter_statements(self, other):
        """Generate ALTER statements to transform the other type to match this one."""
        return other.comment_alter_statements(self)

    @property
    def safer_create_statements(self):
        """Generate both CREATE and COMMENT statements for type creation."""
        statements = [self.create_statement]
        if self.comment:
            statements.append(self.comment_statement)
        return statements


class InspectedDomain(Inspected):
    def __init__(
        self,
        name,
        schema,
        data_type,
        collation,
        constraint_name,
        not_null,
        default,
        check,
    ):
        self.name = name
        self.schema = schema
        self.data_type = data_type
        self.collation = collation
        self.constraint_name = constraint_name
        self.not_null = not_null
        self.default = default
        self.check = check

    @property
    def drop_statement(self):
        return "drop domain {};".format(self.signature)

    @property
    def create_statement(self):
        T = """\
create domain {name}
as {_type}
{collation}{default}{nullable}{check}
"""

        sql = T.format(
            name=self.signature,
            _type=self.data_type,
            collation=self.collation_clause,
            default=self.default_clause,
            check=self.check_clause,
            nullable=self.nullable_clause,
        )

        return sql

    @property
    def check_clause(self):
        if self.check:
            return "{}\n".format(self.check)

        return ""

    @property
    def collation_clause(self):
        if self.collation:
            return "collation {}\n".format(self.collation)

        return ""

    @property
    def default_clause(self):
        if self.default:
            return "default {}\n".format(self.default)

        return ""

    @property
    def nullable_clause(self):
        if self.not_null:
            return "not null\n"
        else:
            return "null\n"

    equality_attributes = (
        "schema name data_type collation default constraint_name not_null check".split()
    )

    def __eq__(self, other):
        try:
            return all(
                [
                    getattr(self, a) == getattr(other, a)
                    for a in self.equality_attributes
                ]
            )
        except AttributeError:
            return False


class InspectedExtension(Inspected):
    def __init__(self, name, schema, version=None, oid=None):
        self.name = name
        self.schema = schema
        self.version = version
        self.oid = oid

    @property
    def drop_statement(self):
        return "drop extension if exists {};".format(self.quoted_name)

    @property
    def create_statement(self):
        if self.version:
            version_clause = f" version '{self.version}'"
        else:
            version_clause = ""

        return "create extension if not exists {} with schema {}{};".format(
            self.quoted_name, self.quoted_schema, version_clause
        )

    @property
    def update_statement(self):
        if not self.version:
            return None
        return "alter extension {} update to '{}';".format(
            self.quoted_name, self.version
        )

    def alter_statements(self, other=None):
        return [self.update_statement]

    def __eq__(self, other):
        equalities = (
            self.name == other.name,
            self.schema == other.schema,
            self.version == other.version,
        )
        return all(equalities)

    def unversioned_copy(self):
        return InspectedExtension(self.name, self.schema)


class InspectedConstraint(Inspected, TableRelated):
    def __init__(
        self,
        name,
        schema,
        constraint_type,
        table_name,
        definition,
        index,
        is_fk=False,
        is_deferrable=False,
        initially_deferred=False,
        is_not_valid=False,
    ):
        self.name = name
        self.schema = schema
        self.constraint_type = constraint_type
        self.table_name = table_name
        self.definition = definition
        self.index = index
        self.is_fk = is_fk

        self.quoted_full_foreign_table_name = None
        self.fk_columns_local = None
        self.fk_columns_foreign = None

        self.is_deferrable = is_deferrable
        self.initially_deferred = initially_deferred
        self.is_not_valid = is_not_valid

    @property
    def drop_statement(self):
        return "alter table {} drop constraint {};".format(
            self.quoted_full_table_name, self.quoted_name
        )

    @property
    def deferrable_subclause(self):
        # [ DEFERRABLE | NOT DEFERRABLE ] [ INITIALLY DEFERRED | INITIALLY IMMEDIATE ]

        if not self.is_deferrable:
            return ""

        else:
            clause = " DEFERRABLE"

            if self.initially_deferred:
                clause += " INITIALLY DEFERRED"

            return clause

    @property
    def create_statement(self):
        return self.get_create_statement(set_not_valid=False)

    def get_create_statement(self, set_not_valid=False):
        if self.index and self.constraint_type != "EXCLUDE":
            using_clause = "{} using index {}{}".format(
                self.constraint_type, self.quoted_name, self.deferrable_subclause
            )
        else:
            using_clause = self.definition

            # Only append NOT VALID if requested and not already NOT VALID
            if set_not_valid and not self.is_not_valid:
                using_clause += " not valid"

        USING = "alter table {} add constraint {} {};"

        return USING.format(self.quoted_full_table_name, self.quoted_name, using_clause)

    @property
    def can_use_not_valid(self):
        return self.constraint_type in ("CHECK", "FOREIGN KEY") and not self.index

    @property
    def validate_statement(self):
        if self.can_use_not_valid:
            VALIDATE = "alter table {} validate constraint {};"
            return VALIDATE.format(self.quoted_full_table_name, self.quoted_name)

    @property
    def safer_create_statements(self):
        if not self.can_use_not_valid:
            return [self.create_statement]

        # If constraint is defined as NOT VALID, respect that and don't validate
        if self.is_not_valid:
            return [self.create_statement]

        return [self.get_create_statement(set_not_valid=True), self.validate_statement]

    @property
    def quoted_full_name(self):
        return "{}.{}.{}".format(
            quoted_identifier(self.schema),
            quoted_identifier(self.table_name),
            quoted_identifier(self.name),
        )

    @property
    def quoted_full_table_name(self):
        return "{}.{}".format(
            quoted_identifier(self.schema), quoted_identifier(self.table_name)
        )

    def __eq__(self, other):
        equalities = (
            self.name == other.name,
            self.schema == other.schema,
            self.table_name == other.table_name,
            self.definition == other.definition,
            self.index == other.index,
            self.is_deferrable == other.is_deferrable,
            self.initially_deferred == other.initially_deferred,
        )
        return all(equalities)


class InspectedPrivilege(Inspected):
    def __init__(self, object_type, schema, name, privilege, target_user):
        self.schema = schema
        self.object_type = object_type
        self.name = name
        self.privilege = privilege.lower()
        self.target_user = target_user

    @property
    def quoted_target_user(self):
        return quoted_identifier(self.target_user)

    @property
    def drop_statement(self):
        return "revoke {} on {} {} from {};".format(
            self.privilege,
            self.object_type,
            self.quoted_full_name,
            self.quoted_target_user,
        )

    @property
    def create_statement(self):
        return "grant {} on {} {} to {};".format(
            self.privilege,
            self.object_type,
            self.quoted_full_name,
            self.quoted_target_user,
        )

    def __eq__(self, other):
        equalities = (
            self.schema == other.schema,
            self.object_type == other.object_type,
            self.name == other.name,
            self.privilege == other.privilege,
            self.target_user == other.target_user,
        )
        return all(equalities)

    @property
    def key(self):
        return self.object_type, self.quoted_full_name, self.target_user, self.privilege


RLS_POLICY_CREATE = """create policy {name}
on {table_name}
as {permissiveness}
for {commandtype_keyword}
to {roleslist}{qual_clause}{withcheck_clause};
"""

COMMANDTYPES = {"*": "all", "r": "select", "a": "insert", "w": "update", "d": "delete"}


class InspectedRowPolicy(Inspected, TableRelated):
    def __init__(
        self, name, schema, table_name, commandtype, permissive, roles, qual, withcheck
    ):
        self.name = name
        self.schema = schema
        self.table_name = table_name
        self.commandtype = commandtype
        self.permissive = permissive
        self.roles = roles
        self.qual = qual
        self.withcheck = withcheck

    @property
    def permissiveness(self):
        return "permissive" if self.permissive else "restrictive"

    @property
    def commandtype_keyword(self):
        return COMMANDTYPES[self.commandtype]

    @property
    def key(self):
        return "{}.{}".format(self.quoted_full_table_name, self.quoted_name)

    @property
    def create_statement(self):
        if self.qual:
            qual_clause = "\nusing ({})".format(self.qual)
        else:
            qual_clause = ""

        if self.withcheck:
            withcheck_clause = "\nwith check ({})".format(self.withcheck)
        else:
            withcheck_clause = ""

        roleslist = ", ".join(self.roles)

        return RLS_POLICY_CREATE.format(
            name=self.quoted_name,
            table_name=self.quoted_full_table_name,
            permissiveness=self.permissiveness,
            commandtype_keyword=self.commandtype_keyword,
            roleslist=roleslist,
            qual_clause=qual_clause,
            withcheck_clause=withcheck_clause,
        )

    @property
    def drop_statement(self):
        return "drop policy {} on {};".format(
            self.quoted_name, self.quoted_full_table_name
        )

    def __eq__(self, other):
        equalities = (
            self.name == self.name,
            self.schema == other.schema,
            self.permissiveness == other.permissiveness,
            self.commandtype == other.commandtype,
            self.permissive == other.permissive,
            self.roles == other.roles,
            self.qual == other.qual,
            self.withcheck == other.withcheck,
        )
        return all(equalities)


PROPS = "schemas relations tables views functions selectables sequences constraints indexes enums extensions privileges collations triggers rlspolicies"


class PostgreSQL:
    def __init__(self, c, include_internal=False):
        if not hasattr(c, "dialect") and hasattr(c, "s"):
            c = c.s.connection()

        try:
            pg_version = c.dialect.server_version_info[0]
            self.connection_type = "sqlalchemy"
        except AttributeError:
            conn = c.connection

            is_psyco3 = hasattr(conn, "pgconn")

            if is_psyco3:
                conn = conn.pgconn
                self.connection_type = "psycopg3"
            else:
                self.connection_type = "psycopg2"

            server_version = conn.server_version

            pg_version = server_version // 10000

        self.pg_version = pg_version
        self.db_version = pg_version

        def processed(q):
            if not include_internal:
                q = q.replace("-- SKIP_INTERNAL", "")
            if self.pg_version >= 11:
                q = q.replace("-- 11_AND_LATER", "")
            else:
                q = q.replace("-- 10_AND_EARLIER", "")

            if self.pg_version >= 15:
                q = q.replace("-- 15_AND_LATER", "")
            else:
                q = q.replace("-- 14_AND_EARLIER", "")

            if self.pg_version >= 17:
                q = q.replace("-- 17_AND_LATER", "")
            else:
                q = q.replace("-- 16_AND_EARLIER", "")

            if 15 <= self.pg_version <= 16:
                q = q.replace("-- 15_AND_16", "")

            q = q.replace("-- EXTOIDS", self.EXTOIDS_QUERY)

            if self.connection_type == "sqlalchemy":
                from sqlalchemy import text

                q = text(q)

            else:
                q = q.replace(r"\:", ":")
            return q

        self.EXTOIDS_QUERY = EXTOIDS_QUERY
        all_relations_query = ALL_RELATIONS_QUERY

        if pg_version >= 12:
            replace = "-- 12_ONLY"
        else:
            replace = "-- PRE_12"

        all_relations_query = all_relations_query.replace(replace, "")
        self.ALL_RELATIONS_QUERY = processed(all_relations_query)
        self.COLLATIONS_QUERY = processed(COLLATIONS_QUERY)
        self.RLSPOLICIES_QUERY = processed(RLSPOLICIES_QUERY)

        self.INDEXES_QUERY = processed(INDEXES_QUERY)
        self.SEQUENCES_QUERY = processed(SEQUENCES_QUERY)
        self.CONSTRAINTS_QUERY = processed(CONSTRAINTS_QUERY)
        self.FUNCTIONS_QUERY = processed(FUNCTIONS_QUERY)
        self.TYPES_QUERY = processed(TYPES_QUERY)
        self.DOMAINS_QUERY = processed(DOMAINS_QUERY)
        self.EXTENSIONS_QUERY = processed(EXTENSIONS_QUERY)
        self.ENUMS_QUERY = processed(ENUMS_QUERY)
        self.DEPS_QUERY = processed(DEPS_QUERY)
        self.SCHEMAS_QUERY = processed(SCHEMAS_QUERY)
        self.PRIVILEGES_QUERY = processed(PRIVILEGES_QUERY)
        self.TRIGGERS_QUERY = processed(TRIGGERS_QUERY)

        self.c = c
        self.load_all()

    def execute(self, *args, **kwargs):
        if not self.connection_type == "sqlalchemy":
            # breakpoint()
            self.c.execute(*args, **kwargs)

            # if result is None:
            rows = self.c.fetchall()
            descr = list(self.c.description)
            column_names = [c.name for c in descr]
            return [AttDict(zip(column_names, row)) for row in rows]
            # else:
            #     return result
        else:
            return self.c.execute(*args, **kwargs)

    def load_all(self):
        self.load_schemas()
        self.load_all_relations()
        self.load_functions()
        self.selectables = od()
        self.selectables.update(self.relations)
        self.selectables.update(self.composite_types)
        self.selectables.update(self.functions)

        self.load_privileges()
        self.load_triggers()
        self.load_collations()
        self.load_rlspolicies()
        self.load_types()
        self.load_domains()

        self.load_deps()
        self.load_deps_all()

    def load_schemas(self):
        q = self.execute(self.SCHEMAS_QUERY)
        schemas = [InspectedSchema(schema=each.schema) for each in q]
        self.schemas = od((schema.schema, schema) for schema in schemas)

    def load_rlspolicies(self):
        if self.pg_version <= 9:
            self.rlspolicies = od()
            return

        q = self.execute(self.RLSPOLICIES_QUERY)

        rlspolicies = [
            InspectedRowPolicy(
                name=p.name,
                schema=p.schema,
                table_name=p.table_name,
                commandtype=p.commandtype,
                permissive=p.permissive,
                roles=p.roles,
                qual=p.qual,
                withcheck=p.withcheck,
            )
            for p in q
        ]

        self.rlspolicies = od((p.key, p) for p in rlspolicies)

    def load_collations(self):
        q = self.execute(self.COLLATIONS_QUERY)

        collations = [
            InspectedCollation(
                schema=i.schema,
                name=i.name,
                provider=i.provider,
                encoding=i.encoding,
                lc_collate=i.lc_collate,
                lc_ctype=i.lc_ctype,
                version=i.version,
                locale=i.locale,
            )
            for i in q
        ]
        self.collations = od((i.quoted_full_name, i) for i in collations)

    def load_privileges(self):
        q = self.execute(self.PRIVILEGES_QUERY)
        privileges = [
            InspectedPrivilege(
                object_type=i.object_type,
                schema=i.schema,
                name=i.name,
                privilege=i.privilege,
                target_user=i.user,
            )
            for i in q
        ]
        self.privileges = od((i.key, i) for i in privileges)

    def load_deps(self):
        q = self.execute(self.DEPS_QUERY)

        self.deps = list(q)

        for dep in self.deps:
            x = quoted_identifier(dep.name, dep.schema, dep.identity_arguments)
            x_dependent_on = quoted_identifier(
                dep.name_dependent_on,
                dep.schema_dependent_on,
                dep.identity_arguments_dependent_on,
            )
            self.selectables[x].dependent_on.append(x_dependent_on)
            self.selectables[x].dependent_on.sort()

            try:
                self.selectables[x_dependent_on].dependents.append(x)
                self.selectables[x_dependent_on].dependents.sort()
            except LookupError:
                pass

        for k, t in self.triggers.items():
            for dep_name in t.dependent_on:
                try:
                    dependency = self.selectables[dep_name]
                except KeyError:
                    continue
                dependency.dependents.append(k)

        for k, r in self.relations.items():
            for kc, c in r.columns.items():
                if c.is_enum:
                    e_sig = c.enum.signature

                    if e_sig in self.enums:
                        r.dependent_on.append(e_sig)
                        c.enum.dependents.append(k)

            if r.parent_table:
                pt = self.relations[r.parent_table]
                r.dependent_on.append(r.parent_table)
                pt.dependents.append(r.signature)

    def get_dependency_by_signature(self, signature):
        things = [self.selectables, self.enums, self.triggers]

        for thing in things:
            try:
                return thing[signature]
            except KeyError:
                continue

    def load_deps_all(self):
        def get_related_for_item(item, att):
            related = [self.get_dependency_by_signature(_) for _ in getattr(item, att)]
            return [item.signature] + [
                _ for d in related for _ in get_related_for_item(d, att)
            ]

        for k, x in self.selectables.items():
            d_all = get_related_for_item(x, "dependent_on")[1:]
            d_all.sort()
            x.dependent_on_all = d_all
            d_all = get_related_for_item(x, "dependents")[1:]
            d_all.sort()
            x.dependents_all = d_all

    def dependency_order(
        self,
        drop_order=False,
        selectables=True,
        triggers=True,
        enums=True,
        include_fk_deps=False,
    ):
        from graphlib import TopologicalSorter

        graph, things = {}, {}

        if enums:
            things.update(self.enums)
        if selectables:
            things.update(self.selectables)
        if triggers:
            things.update(self.triggers)

        for k, x in things.items():
            dependent_on = list(x.dependent_on)

            if k in self.tables and x.parent_table:
                dependent_on.append(x.parent_table)

            graph[k] = list(x.dependent_on)

        if include_fk_deps:
            fk_deps = {}

            for k, x in self.constraints.items():
                if x.is_fk:
                    t, other_t = (
                        x.quoted_full_table_name,
                        x.quoted_full_foreign_table_name,
                    )
                    fk_deps[t] = [other_t]

            graph.update(fk_deps)

        ts = TopologicalSorter(graph)

        ordering = []

        ts.prepare()

        while ts.is_active():
            items = ts.get_ready()

            itemslist = list(items)

            # itemslist.sort()
            ordering += itemslist
            ts.done(*items)

        if drop_order:
            ordering.reverse()
        return ordering

    @property
    def partitioned_tables(self):
        return od((k, v) for k, v in self.tables.items() if v.is_partitioned)

    @property
    def alterable_tables(self):  # ordinary tables and parent tables
        return od((k, v) for k, v in self.tables.items() if v.is_alterable)

    @property
    def data_tables(self):  # ordinary tables and child tables
        return od((k, v) for k, v in self.tables.items() if v.contains_data)

    @property
    def partitioning_child_tables(self):
        return od(
            (k, v) for k, v in self.tables.items() if v.is_partitioning_child_table
        )

    @property
    def tables_using_partitioning(self):
        return od((k, v) for k, v in self.tables.items() if v.uses_partitioning)

    @property
    def tables_not_using_partitioning(self):
        return od((k, v) for k, v in self.tables.items() if not v.uses_partitioning)

    def load_all_relations(self):
        self.tables = od()
        self.views = od()
        self.materialized_views = od()
        self.composite_types = od()

        q = self.execute(self.ENUMS_QUERY)
        enumlist = [
            InspectedEnum(
                name=i.name,
                schema=i.schema,
                elements=i.elements,
                pg_version=self.pg_version,
            )
            for i in q
        ]
        self.enums = od((i.quoted_full_name, i) for i in enumlist)
        q = self.execute(self.ALL_RELATIONS_QUERY)

        for _, g in groupby(q, lambda x: (x.relationtype, x.schema, x.name)):
            clist = list(g)
            f = clist[0]

            def get_enum(name, schema):
                if not name and not schema:
                    return None

                quoted_full_name = "{}.{}".format(
                    quoted_identifier(schema), quoted_identifier(name)
                )

                return self.enums.get(quoted_full_name)

            columns = [
                ColumnInfo(
                    name=c.attname,
                    dbtype=c.datatype,
                    dbtypestr=c.datatypestring,
                    pytype=None,
                    default=c.defaultdef,
                    not_null=c.not_null,
                    is_enum=c.is_enum,
                    enum=get_enum(c.enum_name, c.enum_schema),
                    collation=c.collation,
                    is_identity=c.is_identity,
                    is_identity_always=c.is_identity_always,
                    is_generated=c.is_generated,
                    can_drop_generated=self.pg_version >= 13,
                    can_set_expression=self.pg_version >= 17,
                )
                for c in clist
                if c.position_number
            ]

            s = InspectedSelectable(
                oid=f.oid,
                name=f.name,
                schema=f.schema,
                columns=od((c.name, c) for c in columns),
                relationtype=f.relationtype,
                definition=f.definition,
                comment=f.comment,
                parent_table=f.parent_table,
                partition_def=f.partition_def,
                rowsecurity=f.rowsecurity,
                forcerowsecurity=f.forcerowsecurity,
                persistence=f.persistence,
                options=f.options,
            )
            RELATIONTYPES = {
                "r": "tables",
                "v": "views",
                "m": "materialized_views",
                "c": "composite_types",
                "p": "tables",
            }

            att = getattr(self, RELATIONTYPES[f.relationtype])
            att[s.quoted_full_name] = s

        for k, t in self.tables.items():
            if t.is_inheritance_child_table:
                parent_table = self.tables[t.parent_table]
                for cname, c in t.columns.items():
                    if cname in parent_table.columns:
                        c.is_inherited = True

        self.relations = od()
        for x in (self.tables, self.views, self.materialized_views):
            self.relations.update(x)
        q = self.execute(self.INDEXES_QUERY)

        indexlist = [
            InspectedIndex(
                name=i.name,
                schema=i.schema,
                definition=i.definition,
                table_name=i.table_name,
                key_columns=i.key_columns,
                index_columns=i.index_columns,
                included_columns=i.included_columns,
                key_options=i.key_options,
                num_att=i.num_att,
                is_unique=i.is_unique,
                is_pk=i.is_pk,
                is_exclusion=i.is_exclusion,
                is_immediate=i.is_immediate,
                is_clustered=i.is_clustered,
                key_collations=i.key_collations,
                key_expressions=i.key_expressions,
                partial_predicate=i.partial_predicate,
                algorithm=i.algorithm,
            )
            for i in q
        ]
        self.indexes = od((i.quoted_full_name, i) for i in indexlist)
        q = self.execute(self.SEQUENCES_QUERY)

        sequencelist = [
            InspectedSequence(
                name=i.name,
                schema=i.schema,
                table_name=i.table_name,
                column_name=i.column_name,
            )
            for i in q
        ]
        self.sequences = od((i.quoted_full_name, i) for i in sequencelist)
        q = self.execute(self.CONSTRAINTS_QUERY)

        constraintlist = []

        for i in q:
            constraint = InspectedConstraint(
                name=i.name,
                schema=i.schema,
                constraint_type=i.constraint_type,
                table_name=i.table_name,
                definition=i.definition,
                index=getattr(i, "index"),
                is_fk=i.is_fk,
                is_deferrable=i.is_deferrable,
                initially_deferred=i.initially_deferred,
                is_not_valid=i.is_not_valid,
            )
            if constraint.index:
                index_name = quoted_identifier(constraint.index, schema=i.schema)
                index = self.indexes[index_name]
                index.constraint = constraint
                constraint.index = index

            if constraint.is_fk:
                constraint.quoted_full_foreign_table_name = quoted_identifier(
                    i.foreign_table_name, schema=i.foreign_table_schema
                )
                constraint.fk_columns_foreign = i.fk_columns_foreign
                constraint.fk_columns_local = i.fk_columns_local

            constraintlist.append(constraint)

        self.constraints = od((i.quoted_full_name, i) for i in constraintlist)

        q = self.execute(self.EXTENSIONS_QUERY)
        extensionlist = [
            InspectedExtension(
                name=i.name, schema=i.schema, version=i.version, oid=i.oid
            )
            for i in q
        ]
        # extension names are unique per-database rather than per-schema like other things (even though extensions are assigned to a particular schema)
        self.extensions = od((i.name, i) for i in extensionlist)
        # add indexes and constraints to each table
        for each in self.indexes.values():
            t = each.quoted_full_table_name
            n = each.quoted_full_name

            try:
                self.relations[t].indexes[n] = each
            except Exception:
                pass
        for each in self.constraints.values():
            t = each.quoted_full_table_name
            n = each.quoted_full_name
            self.relations[t].constraints[n] = each

    @property
    def extensions_without_versions(self):
        return {k: v.unversioned_copy() for k, v in self.extensions.items()}

    def load_functions(self):
        self.functions = od()
        q = self.execute(self.FUNCTIONS_QUERY)
        for _, g in groupby(q, lambda x: (x.schema, x.name, x.identity_arguments)):
            clist = list(g)
            f = clist[0]
            outs = [c for c in clist if c.parameter_mode == "OUT"]
            columns = [
                ColumnInfo(
                    name=c.parameter_name,
                    dbtype=c.data_type,
                    pytype=None,  # self.to_pytype(c.data_type),
                )
                for c in outs
            ]
            if outs:
                columns = [
                    ColumnInfo(
                        name=c.parameter_name,
                        dbtype=c.data_type,
                        pytype=None,  # self.to_pytype(c.data_type),
                    )
                    for c in outs
                ]
            else:
                columns = [
                    ColumnInfo(
                        name=f.name,
                        dbtype=f.data_type,
                        pytype=None,  # self.to_pytype(f.returntype),
                        default=f.parameter_default,
                    )
                ]
            plist = [
                ColumnInfo(
                    name=c.parameter_name,
                    dbtype=c.data_type,
                    pytype=None,  # self.to_pytype(c.data_type),
                    default=c.parameter_default,
                )
                for c in clist
                if c.parameter_mode == "IN"
            ]
            s = InspectedFunction(
                schema=f.schema,
                name=f.name,
                columns=od((c.name, c) for c in columns),
                inputs=plist,
                identity_arguments=f.identity_arguments,
                result_string=f.result_string,
                language=f.language,
                definition=f.definition,
                strictness=f.strictness,
                security_type=f.security_type,
                volatility=f.volatility,
                full_definition=f.full_definition,
                comment=f.comment,
                returntype=f.returntype,
                kind=f.kind,
            )

            identity_arguments = "({})".format(s.identity_arguments)
            self.functions[s.quoted_full_name + identity_arguments] = s

    def load_triggers(self):
        q = self.execute(self.TRIGGERS_QUERY)
        triggers = [
            InspectedTrigger(
                i.name,
                i.schema,
                i.table_name,
                i.proc_schema,
                i.proc_name,
                i.enabled,
                i.full_definition,
            )
            for i in q
        ]  # type: list[InspectedTrigger]
        self.triggers = od((t.signature, t) for t in triggers)

    def load_types(self):
        q = self.execute(self.TYPES_QUERY)

        def col(defn):
            return defn["attribute"], defn["type"]

        types = [
            InspectedType(
                i.name, i.schema, dict(col(_) for _ in i.columns), comment=i.comment
            )
            for i in q
        ]  # type: list[InspectedType]
        self.types = od((t.signature, t) for t in types)

    def load_domains(self):
        q = self.execute(self.DOMAINS_QUERY)

        def col(defn):
            return defn["attribute"], defn["type"]

        domains = [
            InspectedDomain(
                i.name,
                i.schema,
                i.data_type,
                i.collation,
                i.constraint_name,
                i.not_null,
                i.default,
                i.check,
            )
            for i in q
        ]  # type: list[InspectedType]
        self.domains = od((t.signature, t) for t in domains)

    def filter_schema(self, schema=None, exclude_schema=None):
        if schema and exclude_schema:
            raise ValueError("Can only have schema or exclude schema, not both")

        def equal_to_schema(x):
            return x.schema == schema

        def not_equal_to_exclude_schema(x):
            return x.schema != exclude_schema

        if schema:
            comparator = equal_to_schema
        elif exclude_schema:
            comparator = not_equal_to_exclude_schema
        else:
            raise ValueError("schema or exclude_schema must be not be none")

        for prop in PROPS.split():
            att = getattr(self, prop)
            filtered = {k: v for k, v in att.items() if comparator(v)}
            setattr(self, prop, filtered)

    def _as_dicts(self):
        done = set()

        def obj_to_d(x, k=None):
            if id(x) in done:
                if isinstance(x, (str, bool, int)):
                    return x
                elif hasattr(x, "quoted_full_name"):
                    return x.quoted_full_name
                return "..."
            done.add(id(x))

            if isinstance(x, dict):
                return {k: obj_to_d(v, k) for k, v in x.items()}

            elif isinstance(x, (ColumnInfo, Inspected)):

                def safe_getattr(x, k):
                    try:
                        return getattr(x, k)
                    except NotImplementedError:
                        return "NOT IMPLEMENTED"

                return {
                    k: obj_to_d(safe_getattr(x, k), k)
                    for k in dir(x)
                    if not k.startswith("_") and not callable(safe_getattr(x, k))
                }
            else:
                return str(x)

        d = {}

        for prop in PROPS.split():
            att = getattr(self, prop)

            _d = {k: obj_to_d(v) for k, v in att.items()}

            d[prop] = _d

        return d

    def encodeable_definition(self):
        return self._as_dicts()

    def as_yaml(self):
        from io import StringIO as sio

        import yaml

        s = sio()

        yaml.safe_dump(self.encodeable_definition(), s)

        return s.getvalue()

    def one_schema(self, schema):
        self.filter_schema(schema=schema)

    def exclude_schema(self, schema):
        self.filter_schema(exclude_schema=schema)

    def __eq__(self, other):
        """
        :type other: PostgreSQL
        :rtype: bool
        """

        return (
            type(self) is type(other)
            and self.schemas == other.schemas
            and self.relations == other.relations
            and self.sequences == other.sequences
            and self.enums == other.enums
            and self.constraints == other.constraints
            and self.extensions == other.extensions
            and self.functions == other.functions
            and self.triggers == other.triggers
            and self.collations == other.collations
            and self.rlspolicies == other.rlspolicies
        )



def format_table_of_dicts(table, hierarchical=False):
    """Format a list of dictionaries as a table string."""
    if not table:
        return "Empty table."

    keys = list(table[0].keys())
    types = [str(type(value).__name__) for value in table[0].values()]

    # Convert all values to strings
    string_table = [{k: str(v) for k, v in row.items()} for row in table]

    # Add header row
    string_table.insert(0, {k: k for k in keys})

    # Calculate column widths
    widths = {key: 0 for key in keys}
    for row in string_table:
        for key in keys:
            length = len(row[key])
            if widths[key] < length:
                widths[key] = length

    # Add separator row
    string_table.insert(1, {k: "-" * widths[k] for k in widths})

    # Apply alignment based on type
    for row in string_table:
        for type_name, key in zip(types, keys):
            width = widths[key]
            value = row[key]

            if type_name == "str":
                row[key] = value.ljust(width)
            else:
                row[key] = value.rjust(width)

    # Apply hierarchical formatting if requested
    if hierarchical:
        previous = None
        for row in string_table:
            row_copy = dict(row)

            if previous:
                for key in keys:
                    if row[key] == previous[key]:
                        row[key] = " " * len(row[key])
                    else:
                        break

            previous = row_copy

    return "\n".join(" ".join(row.values()) for row in string_table)



from .inspector import NullInspector
from .pg import PostgreSQL

SUPPORTED = {"postgresql": PostgreSQL}


def get_inspector(x, schema=None, exclude_schema=None):
    if schema and exclude_schema:
        raise ValueError("Cannot provide both schema and exclude_schema")
    if x is None:
        return NullInspector()

    if hasattr(x, "url") and hasattr(x, "URL"):
        with x.t() as t:
            inspected = PostgreSQL(t.c)
    else:
        try:
            inspected = PostgreSQL(x.c)
        except AttributeError:
            inspected = PostgreSQL(x)

    if schema:
        inspected.one_schema(schema)
    elif exclude_schema:
        inspected.exclude_schema(exclude_schema)
    return inspected



import random
import string
from copy import deepcopy
from dataclasses import make_dataclass

from .formatting import format_table_of_dicts


def rando(length=10) -> str:
    """Generate a random string of lowercase letters."""
    return "".join(random.choice(string.ascii_lowercase) for _ in range(length))


def dedup(rows: list[tuple]):
    """Remove duplicate rows while preserving order."""
    return list({k: None for k in rows}.keys())


class DictRow(dict):
    def __str__(self) -> str:
        indent = 2

        maxlen = max(len(repr(_)) for _ in self) + indent

        items = [f"{repr(k).rjust(maxlen)}: {repr(v)}" for k, v in self.items()]

        mid = ",\n".join(items)

        return "{\n" + mid + "\n}\n"


class Rows(list):
    def __init__(self, *args, **kwargs):
        self.paging = None
        self.column_info = None

        super().__init__(*args, **kwargs)

    def __getitem__(self, x):
        if isinstance(x, (int, slice)):
            return super().__getitem__(x)

        if x not in self.column_info:
            raise ValueError(f"no such column: {x}")

        numbers = {c: i for i, c in enumerate(self.column_info)}

        i = numbers[x]

        return [_[i] for _ in self]

    def keys(self):
        return list(self.column_info)

    @property
    def d(self) -> list[dict]:
        return [DictRow(zip(self.column_info, _)) for _ in self]

    @property
    def dc(self) -> list:
        class_name = f"Row_{rando()}"
        _class = make_dataclass(class_name, list(self.column_info))
        return [_class(**d) for d in self.d]

    def as_dicts(self):
        return self.d

    def __str__(self) -> str:
        return "\n" + format_table_of_dicts(self.as_dicts())

    def __repr__(self) -> str:
        return str(self)

    def __add__(self, other):
        if self.column_info != other.column_info:
            raise ValueError("Cannot add Rows with mismatching columns")
        rows = Rows(list(deepcopy(self)) + list(deepcopy(other)))
        rows.column_info = self.column_info
        return rows

    @property
    def hierarchical(self) -> str:
        return format_table_of_dicts(self.as_dicts(), hierarchical=True)

    @property
    def guessed_sql_columns(self):
        from .typeguess import guess_sql_type_of_values

        return {k: guess_sql_type_of_values(self[k]) for k in self.column_info}

    @classmethod
    def from_dicts(cls, list_of_dicts):
        """Create a Rows object from a list of dictionaries."""
        first = list_of_dicts[0]
        column_names = list(first.keys())

        rows = cls(list(d.values()) for d in list_of_dicts)
        rows.column_info = {k: "text" for k in column_names}

        return rows

    def get_columns(self, clist, dedup=False):
        included_indexes = [i for i, x in enumerate(self.column_info) if x in clist]

        column_info = {k: v for k, v in self.column_info.items() if k in clist}

        if dedup:
            rows = Rows(
                list(
                    {
                        tuple(row[i] for i in included_indexes): None for row in self
                    }.keys()
                )
            )
        else:
            rows = Rows(tuple(row[i] for i in included_indexes) for row in self)

        rows.column_info = column_info
        return rows

    def rename(self, remapping):
        self.column_info = {remapping.get(k, k): v for k, v in self.column_info.items()}



from collections.abc import Mapping
from string import ascii_lowercase, digits

from .rows import Rows
from .syntax import quoted_identifier as qi

INSERT = """
    insert into
        {table} ({colspec})
    values
        [values]
"""


INSERT_DEFAULT = """
    insert into
        {table}
    default values
"""


INSERT_UPSERT = """
    on conflict ({upsertkeyspec})
    do update set
        {upsertspec}
"""


INSERT_UPSERT = """
    on conflict ({upsertkeyspec})
    do update set
        {upsertspec}

"""


INSERT_UPSERT_DO_NOTHING = """
    on conflict ({upsertkeyspec})
    do nothing
"""


ALLOWED = set(ascii_lowercase + digits + "_")
LOWERCASE = set(ascii_lowercase)


def renamed_key(k: str):
    k = k.lower()

    filtered = "".join(_ for _ in k if _ in ALLOWED)

    if not filtered:
        return "k"

    if filtered[0] not in LOWERCASE:
        filtered = f"k{filtered}"

    return filtered


def key_renaming(keylist):
    d = {}

    rekeys = set()

    for k in keylist:
        renamed = renamed_key(k)
        modified = renamed

        i = 0

        while True:
            if modified not in rekeys:
                rekeys.add(modified)

                if k != modified:
                    d[k] = modified
                break
            else:
                modified = modified + f"_{i}"
                i += 1

    return d


class Inserting:
    def insert(self, table, rows, upsert_on=None, returning=None, update_cols=None):
        if isinstance(rows, Mapping):
            rows = [rows]

        if isinstance(upsert_on, str):
            upsert_on = [upsert_on]

        if not rows:
            raise ValueError("empty list of rows, nothing to upsert")

        if returning is None:
            returning = not len(rows) > 1

        if isinstance(rows, Rows):
            rows = rows.as_dicts()

        keys = list(rows[0].keys())

        if renaming_map := key_renaming(keys):
            rows = [{renaming_map[k]: v for k, v in d.items()} for d in rows]

        if keys:
            colspec = ", ".join([qi(k) for k in keys])

            if renaming_map:
                valuespec = ", ".join(f":{renaming_map[k]}" for k in keys)
            else:
                valuespec = ", ".join(f":{k}" for k in keys)
            q = INSERT.format(table=table, colspec=colspec, valuespec=valuespec)
        else:
            q = INSERT_DEFAULT.format(table=table)

        if upsert_on:
            if update_cols is None:
                upsert_keys = list(keys)
            else:
                upsert_keys = list(update_cols)

            for k in upsert_on:
                if k in upsert_keys:
                    upsert_keys.remove(k)

            upsertkeyspec = ", ".join([qi(k) for k in upsert_on])

            if upsert_keys:
                upsertspec = ", ".join(
                    f"{qi(k)} = excluded.{qi(k)}" for k in upsert_keys
                )

                q_upsert = INSERT_UPSERT.format(
                    upsertkeyspec=upsertkeyspec, upsertspec=upsertspec
                )
            else:
                q_upsert = INSERT_UPSERT_DO_NOTHING.format(upsertkeyspec=upsertkeyspec)

            q = q + q_upsert
        if returning:
            q += " returning *"

        return self.qvalues(q, rows, fetch=bool(returning))



# Copyright results Pty Ltd 2022

from itertools import zip_longest

from .rows import Rows

ASC_OR_DESC = ("desc", "asc")


def quoted(column_name):
    """Quote a column name for SQL."""
    return f'"{column_name}"'


def parse_order_by_col(column_spec):
    """Parse a single column specification from ORDER BY clause."""
    column_spec = column_spec.strip()
    tokens = column_spec.rsplit(None, 1)

    if len(tokens) > 1:
        possible_direction = tokens[-1].lower()
        if possible_direction in ASC_OR_DESC:
            column_spec = tokens[0]
            descending = possible_direction == "desc"
        else:
            descending = False
    else:
        descending = False

    column_spec = column_spec.strip().strip('"')
    return column_spec, descending


def parse_order_by(order_by):
    """Parse an ORDER BY clause into column specifications."""
    columns = order_by.split(",")
    return [parse_order_by_col(col) for col in columns]


def reversed_order_by(order_by):
    """Reverse the direction of all columns in an ORDER BY specification."""
    return [(column, not descending) for column, descending in order_by]


def order_by_from_parsed(columns):
    """Convert parsed ORDER BY columns back to SQL string."""
    column_list = [
        quoted(column) + (" desc" if descending else "")
        for column, descending in columns
    ]
    return ", ".join(column_list)


PAGED_QUERY = """
select * from
(
{q}
) unpaged_table
{bookmark}
{order_by}
{limit}
"""

PARAM_PREFIX = "paging_"


def get_paged_query(
    query, params, order_by, bookmark=None, per_page=10, backwards=False
):
    query, paging_query_params = paging_wrapped_query(
        query, order_by, bookmark, per_page, backwards
    )
    params.update(paging_query_params)

    return query, params


def get_paged_rows(rows, paging_params):
    rows.paging = Paging(rows, **paging_params)
    return rows


def get_page(
    cursor,
    query,
    *args,
    bookmark=None,
    order_by,
    per_page,
    backwards,
    **kwargs,
):
    query, page_params = paging_wrapped_query(
        query, order_by, bookmark, per_page, backwards
    )

    argslist = list(args)
    try:
        params = argslist[0]
    except IndexError:
        argslist.append({})
        params = argslist[0]

    params.update(page_params)

    argslist[0] = params
    args = tuple(argslist)

    results, d = cursor.qd(query, *args, **kwargs)

    results = Rows(results)

    return results


def bind_pairs_iter(cols, bookmark, swap_on_descending=False):
    for i, zipped in enumerate(zip(cols, bookmark)):
        col, val = zipped
        name, is_descending = col
        lowercase_name = name.lower()
        bind = f"%({PARAM_PREFIX}{lowercase_name})s"

        if swap_on_descending and is_descending:
            yield name, bind
        else:
            yield bind, name


def make_bookmark_where_clause(cols, bookmark):
    if bookmark is None:
        return ""

    pairslist = bind_pairs_iter(cols, bookmark, swap_on_descending=True)

    b, a = zip(*pairslist)
    if len(a) > 1 or len(b) > 1:
        a, b = ", ".join(a), ", ".join(b)
        return f"where row({a}) > row({b})"

    else:
        return f"where {a[0]} > {b[0]}"


def paging_params(cols, bookmark):
    names = [PARAM_PREFIX + c[0].lower() for c in cols]
    return dict(zip_longest(names, bookmark or []))


def paging_wrapped_query(query, order_by, bookmark, per_page, backwards):
    cols = parse_order_by(order_by)
    if backwards:
        cols = reversed_order_by(cols)

    bookmark_clause = make_bookmark_where_clause(cols, bookmark)
    order_list = order_by_from_parsed(cols)
    order_by = f"order by {order_list}"

    limit = f"limit {per_page + 1}"

    params = paging_params(cols, bookmark)
    formatted = PAGED_QUERY.format(
        q=query, bookmark=bookmark_clause, order_by=order_by, limit=limit
    )
    return formatted, params


class Paging:
    def __init__(
        self, results, *, order_by, per_page=10, bookmark=None, backwards=False
    ):
        self.results = results
        self.per_page = per_page
        self.order_by = order_by
        self.bookmark = bookmark
        self.backwards = backwards
        self.parsed_order_by = parse_order_by(order_by)
        self.order_keys = [c[0] for c in self.parsed_order_by]
        # self.d = d

        try:
            self.discarded_item = results.pop(per_page)
            self.has_more = True
        except IndexError:
            self.discarded_item = None
            self.has_more = False

        if backwards:
            results.reverse()
        self.count = len(self.results)

        if self.count:
            self.names = list(self.results.keys())
            self.names_i = {n: i for i, n in enumerate(self.names)}
        else:
            self.names = None
            self.names_i = None

    @property
    def has_after(self):
        return self.discarded_item is not None

    @property
    def has_before(self):
        return self.bookmark is not None

    @property
    def has_next(self):
        if self.backwards:
            return self.has_before
        else:
            return self.has_after

    @property
    def has_prev(self):
        if self.backwards:
            return self.has_after
        else:
            return self.has_before

    @property
    def at_start(self):
        return not self.has_prev

    @property
    def at_end(self):
        return not self.has_next

    @property
    def is_all(self):
        return self.at_start and self.at_end

    @property
    def is_full(self):
        return self.count == self.per_page

    @property
    def past_start(self):
        return self.backwards and self.at_start and not self.is_full

    @property
    def past_end(self):
        return not self.backwards and self.at_end and not self.is_full

    @property
    def next(self):
        if not self.is_empty:
            return self.get_bookmark(self.results[-1])

    @property
    def prev(self):
        if not self.is_empty:
            return self.get_bookmark(self.results[0])

    @property
    def is_empty(self):
        return not bool(self.count)

    @property
    def current(self):
        return self.bookmark

    @property
    def current_reversed(self):
        if self.discarded_item:
            return self.get_bookmark(self.discarded_item)

    def get_bookmark(self, result_row):
        return tuple(result_row[self.names_i[k]] for k in self.order_keys)



from . import command  # noqa
from .database import db  # noqa
from .loading import rows_from_text
from .psyco import quoted_identifier
from .statements import create_table_statement
from .tempdb import temporary_local_db

from .urls import URL, url

__all__ = [
    "command",
    "db",
    "rows_from_text",
    "quoted_identifier",
    "create_table_statement",
    "temporary_local_db",
    "URL",
    "url",
]

__version__ = "1.4.1769602926"



import inspect
from reprlib import recursive_repr

# from pkg_resources import resource_stream as pkg_resource_stream

from importlib.resources import files


def connection_from_s_or_c(s_or_c):  # pragma: no cover
    if hasattr(s_or_c, "s"):
        s_or_c = s_or_c.s

    try:
        s_or_c.engine
        return s_or_c

    except AttributeError:
        try:
            return s_or_c.connection()
        except (AttributeError, TypeError):
            return s_or_c


class AutoRepr:  # pragma: no cover
    @recursive_repr()
    def __repr__(self):
        done = set()

        cname = self.__class__.__name__

        vals = []
        for k in sorted(dir(self)):
            v = getattr(self, k)

            if not k.startswith("_") and (not callable(v)) and id(v) not in done:
                done.add(id(v))

                attr = "{}={}".format(k, repr(v))

                vals.append(attr)
        return "{}({})".format(cname, ", ".join(vals))

    def __str__(self):
        return repr(self)

    def __ne__(self, other):
        return not self == other


def unquoted_identifier(identifier, *, schema=None, identity_arguments=None):
    if identifier is None and schema is not None:
        return schema
    s = "{}".format(identifier)
    if schema:
        s = "{}.{}".format(schema, s)
    if identity_arguments is not None:
        s = "{}({})".format(s, identity_arguments)
    return s


def quoted_identifier(identifier, schema=None, identity_arguments=None):
    if identifier is None and schema is not None:
        return '"{}"'.format(schema.replace('"', '""'))
    s = '"{}"'.format(identifier.replace('"', '""'))
    if schema:
        s = '"{}".{}'.format(schema.replace('"', '""'), s)
    if identity_arguments is not None:
        s = "{}({})".format(s, identity_arguments)
    return s


def external_caller_package():
    """Get the package name of the external caller.

    Returns the __package__ of the calling module, which is needed for
    importlib.resources.files() compatibility with Python < 3.12.
    In Python 3.11 and earlier, files() requires a package, not a module.
    """
    for frame_info in inspect.stack():
        module = inspect.getmodule(frame_info[0])
        if module is not None and module.__name__ != __name__:
            # Return the package, not the module name
            # For a module like 'results.schemainspect.pg.obj',
            # __package__ is 'results.schemainspect.pg'
            return module.__package__ or module.__name__
    return __name__  # pragma: no cover


def resource_stream(subpath):
    package_name = external_caller_package()
    return files(package_name).joinpath(subpath).open()


def resource_text(subpath):
    with resource_stream(subpath) as f:
        return f.read()



from itertools import zip_longest


def transposed(in_data):
    return list(zip_longest(*in_data))


def rows_to_table(rows, sep=" "):
    rows_out = [sep.join(values) for r in rows for values in r]
    return "\n" + "\n".join(rows_out) + "\n"


def t(rows, sep=" ", markdown=True):
    is_dicts = rows and all(hasattr(rows[0], attr) for attr in "keys values".split())
    if not rows:
        return None

    else:
        if is_dicts:
            keys = list(rows[0].keys())
            rows = [_.values() for _ in rows]
        else:
            keys = rows[0]

    if markdown:
        sep = " | "

    rows = [keys] + rows

    columns = transposed(rows)
    widths = [max(len(str(_)) for _ in c) for c in columns]

    rows_out = [
        sep.join(str(value).ljust(widths[i]) for i, value in enumerate(r)) for r in rows
    ]

    if markdown:

        def spacer(width, centered=False, right=False):
            start = ":" if centered else " "
            end = ":" if centered or right else " "
            return start + "*" * width + end

        rows_out.insert(1, sep.join(["-" * w for w in widths]))

    return "\n".join(rows_out)



from collections import OrderedDict as od

from .misc import AutoRepr, quoted_identifier, unquoted_identifier


class Inspected(AutoRepr):
    @property
    def quoted_full_name(self):
        return quoted_identifier(self.name, schema=self.schema)

    @property
    def signature(self):
        return self.quoted_full_name

    @property
    def unquoted_full_name(self):
        return unquoted_identifier(self.name, schema=self.schema)

    @property
    def quoted_name(self):
        return quoted_identifier(self.name)

    @property
    def quoted_schema(self):
        return quoted_identifier(self.schema)

    def __ne__(self, other):
        return not self == other


class TableRelated(object):
    @property
    def quoted_full_table_name(self):
        return "{}.{}".format(
            quoted_identifier(self.schema), quoted_identifier(self.table_name)
        )


class ColumnInfo(AutoRepr):
    def __init__(
        self,
        name,
        dbtype,
        pytype,
        default=None,
        not_null=False,
        is_enum=False,
        enum=None,
        dbtypestr=None,
        collation=None,
        is_identity=False,
        is_identity_always=False,
        is_generated=False,
        is_inherited=False,
        can_drop_generated=False,
        can_set_expression=False,
    ):
        self.name = name or ""
        self.dbtype = dbtype
        self.dbtypestr = dbtypestr or dbtype
        self.pytype = pytype
        self.default = default or None
        self.not_null = not_null
        self.is_enum = is_enum
        self.enum = enum
        self.collation = collation
        self.is_identity = is_identity
        self.is_identity_always = is_identity_always
        self.is_generated = is_generated
        self.is_inherited = is_inherited
        self.can_drop_generated = can_drop_generated
        self.can_set_expression = can_set_expression

    def __eq__(self, other):
        return (
            self.name == other.name
            and self.dbtype == other.dbtype
            and self.dbtypestr == other.dbtypestr
            and self.default == other.default
            and self.not_null == other.not_null
            and self.enum == other.enum
            and self.collation == other.collation
            and self.is_identity == other.is_identity
            and self.is_identity_always == other.is_identity_always
            and self.is_generated == other.is_generated
            and self.is_inherited == other.is_inherited
        )

    def alter_clauses(self, other):
        # ordering:
        # identify must be dropped before notnull
        # notnull must be added before identity

        clauses = []

        notnull_changed = self.not_null != other.not_null
        notnull_added = notnull_changed and self.not_null
        notnull_dropped = notnull_changed and not self.not_null

        default_changed = self.default != other.default

        # default_added = default_changed and self.default
        # default_dropped = default_changed and not self.default

        identity_changed = (
            self.is_identity != other.is_identity
            or self.is_identity_always != other.is_identity_always
        )

        type_or_collation_changed = (
            self.dbtypestr != other.dbtypestr or self.collation != other.collation
        )

        if default_changed:
            clauses.append(self.alter_default_clause_or_generated(other))

        if notnull_added:
            clauses.append(self.alter_not_null_clause)

        if identity_changed:
            clauses.append(self.alter_identity_clause(other))

        if notnull_dropped:
            clauses.append(self.alter_not_null_clause)

        if type_or_collation_changed:
            if self.is_enum and other.is_enum:
                clauses.append(self.alter_enum_type_clause)
            else:
                clauses.append(self.alter_data_type_clause)

        return clauses

    def change_enum_to_string_statement(self, table_name):
        if self.is_enum:
            return "alter table {} alter column {} set data type varchar using {}::varchar;".format(
                table_name, self.quoted_name, self.quoted_name
            )

        else:
            raise ValueError

    def change_string_to_enum_statement(self, table_name):
        if self.is_enum:
            return (
                "alter table {} alter column {} set data type {} using {}::{};".format(
                    table_name,
                    self.quoted_name,
                    self.dbtypestr,
                    self.quoted_name,
                    self.dbtypestr,
                )
            )
        else:
            raise ValueError

    def change_enum_statement(self, table_name):
        if self.is_enum:
            return "alter table {} alter column {} type {} using {}::text::{};".format(
                table_name,
                self.name,
                self.enum.quoted_full_name,
                self.name,
                self.enum.quoted_full_name,
            )
        else:
            raise ValueError

    def drop_default_statement(self, table_name):
        return "alter table {} alter column {} drop default;".format(
            table_name, self.quoted_name
        )

    def add_default_statement(self, table_name):
        return "alter table {} alter column {} set default {};".format(
            table_name, self.quoted_name, self.default
        )

    def alter_table_statements(self, other, table_name):
        prefix = "alter table {}".format(table_name)
        return ["{} {};".format(prefix, c) for c in self.alter_clauses(other)]

    @property
    def quoted_name(self):
        return quoted_identifier(self.name)

    @property
    def creation_clause(self):
        x = "{} {}".format(self.quoted_name, self.dbtypestr)
        if self.is_identity:
            identity_type = "always" if self.is_identity_always else "by default"
            x += " generated {} as identity".format(identity_type)
        if self.not_null:
            x += " not null"
        if self.is_generated:
            x += " generated always as ({}) stored".format(self.default)
        elif self.default:
            x += " default {}".format(self.default)
        return x

    @property
    def add_column_clause(self):
        return "add column {}{}".format(self.creation_clause, self.collation_subclause)

    @property
    def drop_column_clause(self):
        return "drop column {k}".format(k=self.quoted_name)

    @property
    def alter_not_null_clause(self):
        keyword = "set" if self.not_null else "drop"
        return "alter column {} {} not null".format(self.quoted_name, keyword)

    @property
    def alter_default_clause(self):
        if self.default:
            alter = "alter column {} set default {}".format(
                self.quoted_name, self.default
            )
        else:
            alter = "alter column {} drop default".format(self.quoted_name)
        return alter

    def alter_default_clause_or_generated(self, other):
        if self.default:
            alter = "alter column {} set default {}".format(
                self.quoted_name, self.default
            )
        elif other.is_generated and not self.is_generated:
            alter = "alter column {} drop expression".format(self.quoted_name)
        else:
            alter = "alter column {} drop default".format(self.quoted_name)
        return alter

    @property
    def alter_set_expression_clause(self):
        """Generate ALTER COLUMN ... SET EXPRESSION for PG 17+."""
        return "alter column {} set expression as ({})".format(
            self.quoted_name, self.default
        )

    def alter_identity_clause(self, other):
        if self.is_identity:
            identity_type = "always" if self.is_identity_always else "by default"
            if other.is_identity:
                alter = "alter column {} set generated {}".format(
                    self.quoted_name, identity_type
                )
            else:
                alter = "alter column {} add generated {} as identity".format(
                    self.quoted_name, identity_type
                )
        else:
            alter = "alter column {} drop identity".format(self.quoted_name)
        return alter

    @property
    def collation_subclause(self):
        if self.collation:
            collate = " collate {}".format(quoted_identifier(self.collation))
        else:
            collate = ""
        return collate

    @property
    def alter_data_type_clause(self):
        return "alter column {} set data type {}{} using {}::{}".format(
            self.quoted_name,
            self.dbtypestr,
            self.collation_subclause,
            self.quoted_name,
            self.dbtypestr,
        )

    @property
    def alter_enum_type_clause(self):
        return "alter column {} set data type {}{} using {}::text::{}".format(
            self.quoted_name,
            self.dbtypestr,
            self.collation_subclause,
            self.quoted_name,
            self.dbtypestr,
        )


class InspectedSelectable(Inspected):
    def __init__(
        self,
        name,
        schema,
        columns,
        inputs=None,
        definition=None,
        dependent_on=None,
        dependents=None,
        comment=None,
        relationtype="unknown",
        parent_table=None,
        partition_def=None,
        rowsecurity=False,
        forcerowsecurity=False,
        persistence=None,
        options=None,
        oid=None,
    ):
        self.name = name
        self.schema = schema
        self.inputs = inputs or []
        self.columns = columns
        self.definition = definition
        self.relationtype = relationtype
        self.dependent_on = dependent_on or []
        self.dependents = dependents or []
        self.dependent_on_all = []
        self.dependents_all = []
        self.constraints = od()
        self.indexes = od()
        self.comment = comment
        self.parent_table = parent_table
        self.partition_def = partition_def
        self.rowsecurity = rowsecurity
        self.forcerowsecurity = forcerowsecurity
        self.persistence = persistence
        self.options = options
        self.oid = oid

    def __eq__(self, other):
        equalities = (
            type(self) is type(other),
            self.relationtype == other.relationtype,
            self.name == other.name,
            self.schema == other.schema,
            dict(self.columns) == dict(other.columns),
            self.inputs == other.inputs,
            self.definition == other.definition,
            self.parent_table == other.parent_table,
            self.partition_def == other.partition_def,
            self.rowsecurity == other.rowsecurity,
            self.persistence == other.persistence,
            self.options == other.options,
            self.comment == other.comment,
        )
        return all(equalities)

    @property
    def comment_statement(self):
        """Generate a COMMENT ON statement for this object."""
        if self.comment is None:
            return None

        # Determine the object type based on relationtype
        object_type_map = {
            "r": "TABLE",
            "p": "TABLE",  # partitioned table
            "v": "VIEW",
            "m": "MATERIALIZED VIEW",
            "c": "TYPE",  # composite type
            "f": "FUNCTION",  # this is handled separately in InspectedFunction
        }

        object_type = object_type_map.get(self.relationtype, "TABLE")
        escaped_comment = self.comment.replace("'", "''")

        return (
            f"COMMENT ON {object_type} {self.quoted_full_name} IS '{escaped_comment}';"
        )

    @property
    def drop_comment_statement(self):
        """Generate a statement to drop the comment for this object."""
        object_type_map = {
            "r": "TABLE",
            "p": "TABLE",  # partitioned table
            "v": "VIEW",
            "m": "MATERIALIZED VIEW",
            "c": "TYPE",  # composite type
            "f": "FUNCTION",  # this is handled separately in InspectedFunction
        }

        object_type = object_type_map.get(self.relationtype, "TABLE")
        return f"COMMENT ON {object_type} {self.quoted_full_name} IS NULL;"

    def comment_alter_statements(self, other):
        """Generate statements to alter comments between two versions."""
        statements = []

        if self.comment != other.comment:
            if self.comment is None:
                # Adding a comment
                if other.comment is not None:
                    object_type_map = {
                        "r": "TABLE",
                        "p": "TABLE",  # partitioned table
                        "v": "VIEW",
                        "m": "MATERIALIZED VIEW",
                        "c": "TYPE",  # composite type
                        "f": "FUNCTION",  # this is handled separately in InspectedFunction
                    }

                    object_type = object_type_map.get(other.relationtype, "TABLE")
                    escaped_comment = other.comment.replace("'", "''")
                    statements.append(
                        f"COMMENT ON {object_type} {other.quoted_full_name} IS '{escaped_comment}';"
                    )
            elif other.comment is None:
                # Removing a comment
                statements.append(self.drop_comment_statement)
            else:
                # Changing a comment
                object_type_map = {
                    "r": "TABLE",
                    "p": "TABLE",  # partitioned table
                    "v": "VIEW",
                    "m": "MATERIALIZED VIEW",
                    "c": "TYPE",  # composite type
                    "f": "FUNCTION",  # this is handled separately in InspectedFunction
                }

                object_type = object_type_map.get(other.relationtype, "TABLE")
                escaped_comment = other.comment.replace("'", "''")
                statements.append(
                    f"COMMENT ON {object_type} {other.quoted_full_name} IS '{escaped_comment}';"
                )

        return statements

    def only_comment_changed(self, other):
        """Check if only the comment has changed between two versions of the same object."""
        # Check all attributes except comment
        equalities = (
            type(self) is type(other),
            self.relationtype == other.relationtype,
            self.name == other.name,
            self.schema == other.schema,
            dict(self.columns) == dict(other.columns),
            self.inputs == other.inputs,
            self.definition == other.definition,
            self.parent_table == other.parent_table,
            self.partition_def == other.partition_def,
            self.rowsecurity == other.rowsecurity,
            self.persistence == other.persistence,
            self.options == other.options,
        )
        return all(equalities) and self.comment != other.comment



from . import pg
from .get import get_inspector
from .inspected import ColumnInfo, Inspected
from .inspector import NullInspector
from .pg import PostgreSQL

__all__ = [
    "pg",
    "get_inspector",
    "ColumnInfo",
    "Inspected",
    "NullInspector",
    "PostgreSQL",
]



import string
import textwrap

safe_chars = {k for k in (string.ascii_lowercase + "_")}


def quoted_identifier(
    identifier, schema=None, identity_arguments=None, always_quote=False
):
    def qin(x):  # quote if needed
        if (not always_quote) and all(_ in safe_chars for _ in x):
            return x
        else:
            return f'"{x}"'

    s = qin(identifier.replace('"', '""'))
    if schema:
        s = "{}.{}".format(qin(schema.replace('"', '""')), s)
    if identity_arguments is not None:
        s = "{}({})".format(s, identity_arguments)
    return s


qi = quoted_identifier


def inplace_cte(rows, column_names_and_types: dict[str, str], ctename: str):
    # select * from (values (1,2)) as x(a,b);

    prefix = "p"

    def do_row(i, row):
        parts = [f":{prefix}_{i}_{j}" for j in range(len(row))]
        joined = ", ".join(parts)
        return f"({joined})"

    def get_values_expr(rows):
        return ",\n".join(do_row(i, _) for i, _ in enumerate(rows))

    cols_casts = ", ".join(f"{qi(k)}::{v}" for k, v in column_names_and_types.items())

    values_expr = get_values_expr(rows)

    colnames = ", ".join([qi(k) for k in column_names_and_types])
    sig = f"{ctename}({colnames})"

    sql = f"""\
select {cols_casts}
from 
(values 
{textwrap.indent(values_expr, "  ")}
) 
as {sig}"""

    params = {
        f"{prefix}_{i}_{j}": v for i, row in enumerate(rows) for j, v in enumerate(row)
    }

    return sql, params



import re
import string

SAFE_CHARS = {char: None for char in (string.ascii_lowercase + "_")}


def reformat_bind_params(query, rewrite=True):
    """Reformat bind parameters in a SQL query."""
    pattern = re.compile(r"[^:](:[a-z0-9_]+)")

    parts = []
    remainder = 0

    for match in pattern.finditer(query):
        start, end = match.start(1), match.end(1)

        parts.append(query[remainder:start])

        if rewrite:
            varname = query[start + 1 : end]
            parts.append(f"%({varname})s")
        else:
            parts.append(query[start:end])

        remainder = end

    parts.append(query[remainder:])
    return "".join(parts)


def quoted_identifier(
    identifier, schema=None, identity_arguments=None, always_quote=False
):
    """Quote a SQL identifier if needed."""

    def quote_if_needed(text):
        if always_quote:
            return f'"{text}"'
        if all(char in SAFE_CHARS for char in text):
            return text
        else:
            return f'"{text}"'

    escaped_identifier = quote_if_needed(identifier.replace('"', '""'))

    if schema:
        escaped_schema = quote_if_needed(schema.replace('"', '""'))
        escaped_identifier = f"{escaped_schema}.{escaped_identifier}"

    if identity_arguments is not None:
        escaped_identifier = f"{escaped_identifier}({identity_arguments})"

    return escaped_identifier



from contextlib import contextmanager

from .psyco import quoted_identifier


def start_listening(connection, channels):
    names = [quoted_identifier(each) for each in channels]

    c = connection.cursor()

    for name in names:
        c.execute(f"listen {name};")


class ListenNotify:
    def notifications(self, connection, timeout=5, yield_on_timeout=False):
        """Subscribe to PostgreSQL notifications, and handle them
        in infinite-loop style.

        On an actual message, returns the notification (with .pid,
        .channel, and .payload attributes).

        If you've enabled 'yield_on_timeout', yields None on timeout.
        """

        while True:
            gen = connection.notifies(timeout=timeout)
            try:
                for notify in gen:
                    yield notify

                if yield_on_timeout:
                    yield None
                else:
                    break

            finally:
                gen.close()

    @contextmanager
    def listening_connection(self, channels):
        with self.c_autocommit() as cc:
            if isinstance(channels, str):
                channels = [channels]
            start_listening(cc, channels)
            yield cc



from urllib.parse import _coerce_args, urlsplit

NETLOC_PARTS = ["username", "password", "hostname", "port"]

SCHEMES_USING_NETLOC = [
    "",
    "ftp",
    "http",
    "gopher",
    "nntp",
    "telnet",
    "imap",
    "wais",
    "file",
    "mms",
    "https",
    "shttp",
    "snews",
    "prospero",
    "rtsp",
    "rtspu",
    "rsync",
    "svn",
    "svn+ssh",
    "sftp",
    "nfs",
    "git",
    "git+ssh",
    "ws",
    "wss",
]


def urlunsplit(components):
    """Combine the elements of a tuple as returned by urlsplit() into a
    complete URL as a string. The data argument can be any five-item iterable.
    This may result in a slightly different, but equivalent URL, if the URL that
    was parsed originally had unnecessary delimiters (for example, a ? with an
    empty query; the RFC states that these are equivalent)."""
    scheme, netloc, url, query, fragment, _coerce_result = _coerce_args(*components)

    if netloc or (scheme and url[:2] != "//"):
        if url and url[:1] != "/":
            url = "/" + url
        url = "//" + (netloc or "") + url
    if scheme:
        url = scheme + ":" + url
    if query:
        url = url + "?" + query
    if fragment:
        url = url + "#" + fragment
    return _coerce_result(url)


def combine_netloc_parts(parts):
    """Combine netloc parts into a netloc string."""
    username = parts["username"]
    password = parts["password"]
    hostname = parts["hostname"]
    port = parts["port"]

    login = username or ""

    if password:
        login = f"{login}:{password}"

    host_port = hostname or ""

    if host_port:
        if port:
            host_port = f"{host_port}:{port}"

    if login:
        host_port = f"{login}@{host_port}"

    return host_port


class URL:
    """A URL class that allows easy manipulation of URL components."""

    def __init__(self, url_string):
        self.s = url_string
        self.parts = self.urlsplit()

    def urlsplit(self):
        """Split the URL into components."""
        return urlsplit(self.s)

    def __getattr__(self, name):
        if hasattr(self.parts, name):
            return getattr(self.parts, name)

        if hasattr(self, name):
            return getattr(self, name)

        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'"
        )

    def netloc_parts(self):
        """Get netloc parts as a dictionary."""
        return {key: getattr(self.parts, key) for key in NETLOC_PARTS}

    def replace_netloc_part(self, name, value):
        """Replace a specific netloc part."""
        parts = self.netloc_parts()
        parts[name] = value
        combined = combine_netloc_parts(parts)
        self.replace_url_part("netloc", combined)

    def replace_url_part(self, name, value):
        """Replace a specific URL part."""
        self.parts = self.parts._replace(**{name: value})
        self.s = urlunsplit(self.parts)

    def __setattr__(self, name, value):
        if name in ["s", "parts"]:
            super().__setattr__(name, value)
            return

        if name in NETLOC_PARTS:
            self.replace_netloc_part(name, value)
        elif name in self.parts._fields:
            self.replace_url_part(name, value)
        else:
            super().__setattr__(name, value)

    @property
    def absolute_path(self):
        """Get the absolute path (with leading slash)."""
        path = self.path
        if not path.startswith("/"):
            path = "/" + self.path
        return path

    @property
    def relative_path(self):
        """Get the relative path (without leading slash)."""
        return self.path.lstrip("/")

    def __str__(self):
        return urlunsplit(self.parts)


def url(url_string):
    """Create a URL object from a string."""
    return URL(url_string)



import atexit
import getpass
from contextlib import contextmanager
from inspect import currentframe
from pathlib import Path
from threading import get_ident
import multiprocessing

from psycopg import connect as pgconnect
from psycopg_pool import ConnectionPool

from .constants import type_codes
from .createdrop import DatabaseCreateDrop
from .inserting import Inserting
from .notify import ListenNotify
from .paging import get_paged_query, get_paged_rows
from .psyco import reformat_bind_params
from .rows import Rows
from .schemas import Schemas
from .urls import URL

conns = {}


def pool_shutdown():
    for pool in conns.values():
        pool.close()


def close_pool(db_url):
    """Close and remove the connection pool for a specific database URL.

    This should be called before dropping a database to prevent
    "server closed the connection unexpectedly" errors.
    """
    if db_url in conns:
        pool = conns[db_url]
        pool.close()
        del conns[db_url]


atexit.register(pool_shutdown)


def connection_check(conn):
    with conn.cursor() as curs:
        curs.execute("SELECT 1")


def reconnect_failed(connection_pool, **kwargs):
    connection_pool.queue.put("failed")


@contextmanager
def pooled_connection(db_url):
    if db_url not in conns:

        def configure_check(connection, **kwargs):
            pool.queue.put("ok")

        pool = ConnectionPool(
            db_url,
            min_size=1,
            max_size=1024,
            open=False,
            configure=configure_check,
            check=connection_check,
            reconnect_timeout=0,
            reconnect_failed=reconnect_failed,
        )

        queue = multiprocessing.Queue()

        pool.queue = queue

        pool.open()

        conns[db_url] = pool
        result = pool.queue.get()

        if result != "ok":
            raise ValueError("Connection failed")

    else:
        pool = conns[db_url]

    with pool.connection() as conn:
        yield conn


def put_conn(conn, db_url):
    tid = get_ident()

    pool = conns[db_url]
    pool.putconn(conn, tid)


def get_inspected(t, **kwargs):
    from results.schemainspect import get_inspector

    return get_inspector(t.c, **kwargs)


class Transaction(Inserting):
    def __init__(self):
        self._back = 0

    def inspect(self, **kwargs):
        return get_inspected(self, **kwargs)

    def ex(self, *args, **kwargs):
        self.c.execute(*args, **kwargs)

    def execute(self, *args, **kwargs) -> Rows:
        self.c.execute(*args, **kwargs)

        if self.c.description is None:
            return None

        fetched = self.c.fetchall()
        descr = list(self.c.description)

        rows = Rows(fetched)

        rows.paging = None

        column_info = {c.name: type_codes.get(c.type_code, "unknown") for c in descr}

        rows.column_info = column_info

        return rows

    def executemany(self, *args, **kwargs) -> Rows:
        self.c.executemany(*args, **kwargs, returning=True)

        if not self.c.description:
            return None

        fetched = []

        while True:
            for row in self.c.fetchall():
                fetched.append(row)
            if not self.c.nextset():
                break

        if self.c.description is None:
            return None

        descr = list(self.c.description)

        rows = Rows(fetched)

        rows.paging = None

        column_info = {c.name: type_codes.get(c.type_code, "unknown") for c in descr}

        rows.column_info = column_info

        return rows

    @property
    def db_version(self) -> int:
        return self.c.connection.pgconn.server_version // 10000

    def qvalues(self, query, values, fetch=True):
        keylist = list(values[0].keys())

        keylist = [f":{_}" for _ in keylist]

        param_string = f"({', '.join(k for k in keylist)})"

        template = reformat_bind_params(param_string)

        query = query.replace("[values]", template)

        executed = self.executemany(query, values)

        return executed

    def pages(self, *args, paging: dict, **kwargs):
        bookmark = None

        while True:
            paging["bookmark"] = bookmark
            page = self.q(*args, paging=paging, **kwargs)

            if not page:
                break

            yield page

            if page.paging.has_next:
                bookmark = page.paging.next
            else:
                break

    def all_pages(self, *args, paging: dict, **kwargs):
        return list(self.pages(*args, paging=paging, **kwargs))

    def q(
        self,
        query,
        params=None,
        *,
        context=False,
        paging=None,
        fail_on_empty=False,
        colon_bind_params=True,
        **kwargs,
    ):
        if not query.strip():
            if fail_on_empty:
                raise ValueError("Error: Empty query.")
            else:
                return None

        query = query.replace("%", "%" + "%")

        if context:
            frame = currentframe()

        try:
            if context:
                if self._back:
                    fback = frame.f_back.f_back
                else:
                    fback = frame.f_back

                caller_locals = fback.f_locals
                _params = caller_locals
            else:
                _params = {}

            if params:
                _params.update(params)

            if kwargs:
                _params.update(kwargs)

            if _params and colon_bind_params:
                query = reformat_bind_params(query)

            if paging:
                query, _params = get_paged_query(query, _params, **paging)

            rows = self.execute(query, _params)

            if paging:
                rows = get_paged_rows(rows, paging)

            return rows
        finally:
            if context:
                del frame

    def q_from_file(self, path, *args, **kwargs):
        query = Path(path).expanduser().resolve().read_text()
        return self.q(query, *args, **kwargs)

    def notify(self, channel, payload):
        self.q(
            """select pg_notify(:channel, :payload)""",
            dict(channel=channel, payload=payload),
        )


@contextmanager
def autocommit_transaction(db_url):
    conn = pgconnect(db_url)

    conn.autocommit = True

    try:
        with conn.cursor() as curs:
            t = Transaction()
            t.c = curs
            yield t
    finally:
        conn.close()


@contextmanager
def autocommit_connection(db_url):
    conn = pgconnect(db_url)

    conn.autocommit = True

    try:
        yield conn
    finally:
        conn.close()


@contextmanager
def transaction(db_url):  # , cursor_factory=DictCursor):
    with pooled_connection(db_url) as conn:
        with conn.cursor() as curs:
            t = Transaction()
            t.c = curs
            yield t


def db(url=None):
    return Database(url)


class Database(DatabaseCreateDrop, Schemas, ListenNotify):
    def __init__(self, url=None):
        if url is None:
            url = getpass.getuser()

        _url = URL(url)

        if not _url.scheme or _url.scheme == "postgres":
            _url.scheme = "postgresql"

        self.URL = _url
        self.url = str(_url)

    @property
    def url_object(self):
        return URL(self.url)

    def sibling(self, name, allow_self=False):
        sibling_url = URL(self.url)
        sibling_url.path = name

        is_self = sibling_url.path == self.url_object.path

        if is_self and allow_self is False:
            raise ValueError("sibling must not be the same database")
        return db(str(sibling_url))

    @property
    def name(self):
        return self.url_object.relative_path

    @contextmanager
    def t(self):
        with transaction(self.url) as t:
            yield t

    @contextmanager
    def t_autocommit(self):
        with autocommit_transaction(self.url) as t:
            yield t

    @contextmanager
    def c_autocommit(self):
        with autocommit_connection(self.url) as t:
            yield t

    def autocommit(self, *args, **kwargs):
        with self.t_autocommit() as t:
            t.q(*args, **kwargs)

    def __getattr__(self, name):
        def method(*args, **kwargs):
            with self.t() as t:
                t._back = 1
                m = getattr(t, name)
                return m(*args, **kwargs)

        return method

    def __repr__(self):
        return f"db(url={self.url})"



from contextlib import contextmanager

from sqlalchemy import create_engine, text
from sqlalchemy.orm import Session


class SqlaSession:
    """A wrapper around SQLAlchemy Session for simplified usage."""

    def __init__(self, session: Session) -> None:
        self.s: Session = session

    def execute(self, *args, **kwargs):
        """Execute a SQL statement with automatic text conversion."""
        return self.s.execute(text(args[0]), *args[1:], **kwargs)


@contextmanager
def S(url):
    """Create a SQLAlchemy session context manager."""
    if hasattr(url, "url"):
        url = url.url

    # Create an Engine, which the Session will use for connection resources
    engine = create_engine(url)

    try:
        # Create session and add objects
        with Session(engine) as session:
            yield SqlaSession(session)
            session.commit()
    finally:
        # Dispose of the engine to close all connections in the pool
        # This prevents "server closed the connection unexpectedly" errors
        # when temporary databases are dropped
        engine.dispose()



from textwrap import indent


def create_table_statement(name, columns, extra=None, *, uuid_pk=None):
    """Generate a CREATE TABLE SQL statement."""
    extra = extra or {}

    # Build column specifications
    colspec = [
        dict(name=col_name, coltype=col_type, extra=extra.get(col_name))
        for col_name, col_type in columns.items()
    ]

    # Add UUID primary key if requested
    if uuid_pk:
        uuid_col = dict(
            name=uuid_pk,
            coltype="uuid",
            extra="default gen_random_uuid() primary key",
        )
        colspec = [uuid_col] + colspec

    def format_column_line(column_dict):
        """Format a single column definition line."""
        line = f"{column_dict['name']} {column_dict['coltype']}"

        if column_extra := column_dict.get("extra"):
            line += f" {column_extra}"

        return line

    column_lines = [format_column_line(col) for col in colspec]
    column_text = ",\n".join(column_lines)
    column_text = indent(column_text, "  ")

    statement = f"""\
create table {name} (
{column_text}
);

"""
    return statement
